# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import copy
from dataclasses import dataclass, field
from pathlib import Path
from .models import (
    Playbook,
    Play,
    Role,
    Collection,
    TaskFile,
    Task,
    TaskCall,
    InventoryType,
    Object,
    CallObject,
    Variable,
    VariableType,
    BecomeInfo,
    immutable_var_types,
)

p = Path(__file__).resolve().parent
ansible_special_variables = [line.replace("\n", "") for line in open(p / "ansible_variables.txt", "r").read().splitlines()]
_special_var_value = "__ansible_special_variable__"
variable_block_re = re.compile(r"{{[^}]+}}")


def get_object(json_path, type, name, cache={}):
    json_type = ""
    json_str = ""
    cached = cache.get(json_path, None)
    if cached is None:
        if not os.path.exists(json_path):
            raise ValueError('the json path "{}" not found'.format(json_path))
        basename = os.path.basename(json_path)

        if basename.startswith("role-"):
            json_type = "role"
            with open(json_path, "r") as file:
                json_str = file.read()
        elif basename.startswith("collection-"):
            json_type = "collection"
            with open(json_path, "r") as file:
                json_str = file.read()
        if json_str == "":
            raise ValueError("json data is empty")
        cache.update({json_path: (json_type, json_str)})
    else:
        json_type, json_str = cached[0], cached[1]
    if json_type == "role":
        r = Role.from_json(json_str)
        if type == "collection":
            raise ValueError("collection cannot be gotten in a role")
        if type == "role":
            return r
        elif type == "playbook":
            raise ValueError("playbook cannot be gotten in a role")
        elif type == "taskfile":
            for tf in r.taskfiles:
                if tf.defined_in == name:
                    return tf
        elif type == "task":
            for tf in r.taskfiles:
                for t in tf.tasks:
                    if t.id == name:
                        return t
        elif type == "module":
            for m in r.modules:
                if m.fqcn == name:
                    return m
        return None
    elif json_type == "collection":
        c = Collection()
        c.from_json(json_str)
        if type == "collection":
            return c
        if type == "role":
            for r in c.roles:
                if r.fqcn == name:
                    return r
        elif type == "playbook":
            for p in c.playbooks:
                if p.defined_in == name:
                    return p
        elif type == "taskfile":
            for tf in c.taskfiles:
                if tf.defined_in == name:
                    return tf
            for r in c.roles:
                for tf in r.taskfiles:
                    if tf.defined_in == name:
                        return tf
        elif type == "task":
            for p in c.playbooks:
                for t in p.tasks:
                    if t.id == name:
                        return t
            for tf in c.taskfiles:
                for t in tf.tasks:
                    if t.id == name:
                        return t
            for r in c.roles:
                for tf in r.taskfiles:
                    for t in tf.tasks:
                        if t.id == name:
                            return t
        elif type == "module":
            for m in c.modules:
                if m.fqcn == name:
                    return m
        return None
    return None


def recursive_find_variable(var_name, var_dict={}):
    def _visitor(vname, nname, node):
        if nname == vname:
            return node
        if isinstance(node, dict):
            if nname in node:
                return node[nname]
            for k, v in node.items():
                nname2 = "{}.{}".format(nname, k) if nname != "" else k
                if vname.startswith("{}.".format(nname2)):
                    vname2 = vname[len(nname) + 1 :]
                    return _visitor(vname2, nname2, v)
            return None
        else:
            return None

    return _visitor(var_name, "", var_dict)


def flatten(var_dict: dict = {}, _prefix: str = ""):
    flat_vars = {}
    for k, v in var_dict.items():
        if isinstance(v, dict):
            new_prefix = f"{k}." if _prefix == "" else f"{_prefix}{k}"
            sub_flat_vars = flatten(v, new_prefix)
            flat_vars.update(sub_flat_vars)
        else:
            flat_key = f"{_prefix}{k}"
            flat_vars.update({flat_key: v})
    return flat_vars


@dataclass
class Context:
    keep_obj: bool = False
    chain: list = field(default_factory=list)
    variables: dict = field(default_factory=dict)
    options: dict = field(default_factory=dict)
    inventories: list = field(default_factory=list)
    role_defaults: list = field(default_factory=list)
    role_vars: list = field(default_factory=list)
    registered_vars: list = field(default_factory=list)
    set_facts: list = field(default_factory=list)
    task_vars: list = field(default_factory=list)

    become: BecomeInfo = None

    var_set_history: dict = field(default_factory=dict)
    var_use_history: dict = field(default_factory=dict)

    _flat_vars: dict = field(default_factory=dict)

    def add(self, obj, depth_lvl=0):
        _obj = None
        _spec = None
        if isinstance(obj, Object):
            _obj = obj
            _spec = obj
        elif isinstance(obj, CallObject):
            _obj = obj
            _spec = obj.spec
        # variables
        if isinstance(_spec, Playbook):
            self.variables.update(_spec.variables)
            self.update_flat_vars(_spec.variables)
            for key, val in _spec.variables.items():
                current = self.var_set_history.get(key, [])
                current.append(Variable(name=key, value=val, type=VariableType.PlaybookGroupVarsAll, setter=_spec.key))
                self.var_set_history[key] = current
        elif isinstance(_spec, Play):
            self.variables.update(_spec.variables)
            self.update_flat_vars(_spec.variables)
            for key, val in _spec.variables.items():
                current = self.var_set_history.get(key, [])
                current.append(Variable(name=key, value=val, type=VariableType.PlayVars, setter=_spec.key))
                self.var_set_history[key] = current
            if _spec.become:
                self.become = _spec.become
        elif isinstance(_spec, Role):
            self.variables.update(_spec.default_variables)
            self.update_flat_vars(_spec.default_variables)
            self.variables.update(_spec.variables)
            self.update_flat_vars(_spec.variables)
            for var_name in _spec.default_variables:
                self.role_defaults.append(var_name)
            for var_name in _spec.variables:
                self.role_vars.append(var_name)
            for key, val in _spec.default_variables.items():
                current = self.var_set_history.get(key, [])
                current.append(Variable(name=key, value=val, type=VariableType.RoleDefaults, setter=_spec.key))
                self.var_set_history[key] = current
            for key, val in _spec.variables.items():
                current = self.var_set_history.get(key, [])
                current.append(Variable(name=key, value=val, type=VariableType.RoleVars, setter=_spec.key))
                self.var_set_history[key] = current
        elif isinstance(_spec, Collection):
            self.variables.update(_spec.variables)
            self.update_flat_vars(_spec.variables)
        elif isinstance(_spec, TaskFile):
            self.variables.update(_spec.variables)
            self.update_flat_vars(_spec.variables)
        elif isinstance(_spec, Task):
            self.variables.update(_spec.variables)
            self.update_flat_vars(_spec.variables)
            self.variables.update(_spec.registered_variables)
            self.update_flat_vars(_spec.registered_variables)
            self.variables.update(_spec.set_facts)
            self.update_flat_vars(_spec.set_facts)
            for var_name in _spec.registered_variables:
                self.registered_vars.append(var_name)
            for var_name in _spec.set_facts:
                self.set_facts.append(var_name)
            for key, val in _spec.variables.items():
                current = self.var_set_history.get(key, [])
                current.append(Variable(name=key, value=val, type=VariableType.TaskVars, setter=_spec.key))
                self.var_set_history[key] = current
            for key, val in _spec.registered_variables.items():
                current = self.var_set_history.get(key, [])
                current.append(Variable(name=key, value=val, type=VariableType.RegisteredVars, setter=_spec.key))
                self.var_set_history[key] = current
            for key, val in _spec.set_facts.items():
                current = self.var_set_history.get(key, [])
                current.append(Variable(name=key, value=val, type=VariableType.SetFacts, setter=_spec.key))
                self.var_set_history[key] = current
            if _spec.become:
                self.become = _spec.become
        else:
            # Module
            return
        self.options.update(_spec.options)
        chain_node = {"key": _obj.key, "depth": depth_lvl}
        if self.keep_obj:
            chain_node["obj"] = _obj
        self.chain.append(chain_node)

    def resolve_variable(self, var_name, resolve_history={}):
        if var_name in resolve_history:
            val = resolve_history[var_name].get("value", None)
            v_type = resolve_history[var_name].get("type", VariableType.Unknown)
            return val, v_type, resolve_history

        _resolve_history = resolve_history.copy()

        v_type = None
        if var_name in ansible_special_variables:
            v_type = VariableType.HostFacts
            return None, v_type, resolve_history

        if var_name in self.role_vars:
            v_type = VariableType.RoleVars
        elif var_name in self.role_defaults:
            v_type = VariableType.RoleDefaults
        elif var_name in self.registered_vars:
            v_type = VariableType.RegisteredVars
        elif var_name in self.set_facts:
            v_type = VariableType.SetFacts
        else:
            v_type = VariableType.TaskVars

        val = self.variables.get(var_name, None)
        if val is not None:
            _resolve_history[var_name] = {"value": val, "type": v_type}

            if isinstance(val, str):
                resolved_val, _resolve_history = self.resolve_single_variable(val, _resolve_history)
                return resolved_val, v_type, _resolve_history
            elif isinstance(val, list):
                resolved_val_list = []
                for vi in val:
                    resolved_val, _resolve_history = self.resolve_single_variable(vi, _resolve_history)
                    resolved_val_list.append(resolved_val)
                return resolved_val_list, v_type, _resolve_history
            else:
                return val, v_type, _resolve_history

        val = self._flat_vars.get(var_name, None)
        if val is not None:
            _resolve_history[var_name] = {"value": val, "type": v_type}

            if isinstance(val, str):
                resolved_val, _resolve_history = self.resolve_single_variable(val, _resolve_history)
                return resolved_val, v_type, _resolve_history
            elif isinstance(val, list):
                resolved_val_list = []
                for vi in val:
                    resolved_val, _resolve_history = self.resolve_single_variable(vi, _resolve_history)
                    resolved_val_list.append(resolved_val)
                return resolved_val_list, v_type, _resolve_history
            else:
                return val, v_type, _resolve_history

        # TODO: consider group
        inventory_for_all = [iv for iv in self.inventories if iv.inventory_type == InventoryType.GROUP_VARS_TYPE and iv.name == "all"]
        for iv in inventory_for_all:
            iv_var_dict = flatten(iv.variables)
            val = iv_var_dict.get(var_name, None)

            if val is not None:
                _resolve_history[var_name] = {"value": val, "type": v_type}
                v_type = VariableType.InventoryGroupVarsAll
                if isinstance(val, str):
                    resolved_val, _resolve_history = self.resolve_single_variable(val, _resolve_history)
                    return resolved_val, v_type, _resolve_history
                elif isinstance(val, list):
                    resolved_val_list = []
                    for vi in val:
                        resolved_val, _resolve_history = self.resolve_single_variable(vi, _resolve_history)
                        resolved_val_list.append(resolved_val)
                    return resolved_val_list, v_type, _resolve_history
                else:
                    return val, v_type, _resolve_history

        _resolve_history[var_name] = {"value": None, "type": VariableType.Unknown}

        return None, VariableType.Unknown, _resolve_history

    def resolve_single_variable(self, txt, resolve_history=[]):
        new_history = resolve_history.copy()
        if not isinstance(txt, str):
            return txt, new_history
        if "{{" in txt:
            var_names_in_txt = extract_variable_names(txt)
            if len(var_names_in_txt) == 0:
                return txt, new_history
            resolved_txt = txt
            for var_name_in_txt in var_names_in_txt:
                original_block = var_name_in_txt.get("original", "")
                var_name = var_name_in_txt.get("name", "")
                default_var_name = var_name_in_txt.get("default", "")
                var_val_in_txt, _, new_history = self.resolve_variable(var_name, new_history)
                if var_val_in_txt is None and default_var_name != "":
                    var_val_in_txt, _, new_history = self.resolve_variable(default_var_name, new_history)
                if var_val_in_txt is None:
                    return resolved_txt, new_history
                if txt == original_block:
                    return var_val_in_txt, new_history
                resolved_txt = resolved_txt.replace(original_block, str(var_val_in_txt))
            return resolved_txt, new_history
        else:
            return txt, new_history

    def update_flat_vars(self, new_vars: dict, _prefix: str = ""):
        for k, v in new_vars.items():
            if isinstance(v, dict):
                new_prefix = f"{k}." if _prefix == "" else f"{_prefix}{k}"
                self.update_flat_vars(v, new_prefix)
            else:
                flat_key = f"{_prefix}{k}"
                self._flat_vars.update({flat_key: v})
        return

    def chain_str(self):
        lines = []
        for chain_item in self.chain:
            obj = chain_item.get("obj", None)
            depth = chain_item.get("depth", 0)
            indent = "  " * depth
            obj_type = type(obj).__name__
            obj_name = obj.name
            line = "{}{}: {}\n".format(indent, obj_type, obj_name)
            if obj_type == "Task":
                module_name = obj.module
                line = "{}{}: {} (module: {})\n".format(indent, obj_type, obj_name, module_name)
            lines.append(line)
        return "".join(lines)

    def copy(self):
        return Context(
            keep_obj=self.keep_obj,
            chain=copy.copy(self.chain),
            variables=copy.copy(self.variables),
            options=copy.copy(self.options),
            inventories=copy.copy(self.inventories),
            role_defaults=copy.copy(self.role_defaults),
            role_vars=copy.copy(self.role_vars),
            registered_vars=copy.copy(self.registered_vars),
        )
        # return copy.deepcopy(self)


def resolved_vars_contains(resolved_vars, new_var):
    if not isinstance(new_var, dict):
        return False
    new_var_key = new_var.get("key", "")
    if new_var_key == "":
        return False
    if not isinstance(resolved_vars, list):
        return False
    for var in resolved_vars:
        if not isinstance(var, dict):
            continue
        var_key = var.get("key", "")
        if var_key == "":
            continue
        if var_key == new_var_key:
            return True
    return False


def resolve_module_options(context: Context, taskcall: TaskCall):
    resolved_vars = []
    variables_in_loop = []
    used_variables = {}
    if len(taskcall.spec.loop) == 0:
        variables_in_loop = [{}]
    else:
        loop_key = list(taskcall.spec.loop.keys())[0]
        loop_values = taskcall.spec.loop.get(loop_key, [])
        new_var = {
            "key": loop_key,
            "value": loop_values,
            "type": VariableType.LoopVars,
        }
        if not resolved_vars_contains(resolved_vars, new_var):
            resolved_vars.append(new_var)
        if isinstance(loop_values, str):
            var_names = extract_variable_names(loop_values)
            if len(var_names) == 0:
                variables_in_loop.append({loop_key: loop_values})
            else:
                var_name = var_names[0].get("name", "")
                resolved_vars_in_item, v_type, resolve_history = context.resolve_variable(var_name)
                used_variables[var_name] = resolve_history
                new_var = {
                    "key": var_name,
                    "value": resolved_vars_in_item,
                    "type": v_type,
                }
                if not resolved_vars_contains(resolved_vars, new_var):
                    resolved_vars.append(new_var)
                if isinstance(resolved_vars_in_item, list):
                    for vi in resolved_vars_in_item:
                        variables_in_loop.append(
                            {
                                loop_key: vi,
                                "__v_type__": v_type,
                                "__v_name__": var_name,
                            }
                        )
                if isinstance(resolved_vars_in_item, dict):
                    for vi_key, vi_value in resolved_vars_in_item.items():
                        variables_in_loop.append(
                            {
                                loop_key + ".key": vi_key,
                                loop_key + ".value": vi_value,
                                "__v_type__": v_type,
                            }
                        )
                else:
                    variables_in_loop.append(
                        {
                            loop_key: resolved_vars_in_item,
                            "__v_type__": v_type,
                            "__v_name__": var_name,
                        }
                    )
        elif isinstance(loop_values, list):
            for v in loop_values:
                if isinstance(v, str) and variable_block_re.search(v):
                    var_names = extract_variable_names(v)
                    if len(var_names) == 0:
                        variables_in_loop.append({loop_key: v})
                        continue
                    var_name = var_names[0].get("name", "")
                    resolved_vars_in_item, v_type, resolve_history = context.resolve_variable(var_name)
                    used_variables[var_name] = resolve_history
                    new_var = {
                        "key": var_name,
                        "value": resolved_vars_in_item,
                        "type": v_type,
                    }
                    if not resolved_vars_contains(resolved_vars, new_var):
                        resolved_vars.append(new_var)
                    if not isinstance(resolved_vars_in_item, list):
                        variables_in_loop.append(
                            {
                                loop_key: resolved_vars_in_item,
                                "__v_type__": v_type,
                                "__v_name__": var_name,
                            }
                        )
                        continue
                    for vi in resolved_vars_in_item:
                        variables_in_loop.append(
                            {
                                loop_key: vi,
                                "__v_type__": v_type,
                                "__v_name__": var_name,
                            }
                        )
                else:
                    if isinstance(v, dict):
                        tmp_variables = {}
                        for k2, v2 in v.items():
                            key = "{}.{}".format(loop_key, k2)
                            tmp_variables.update({key: v2})
                        variables_in_loop.append(tmp_variables)
                    else:
                        variables_in_loop.append({loop_key: v})
        elif isinstance(loop_values, dict):
            tmp_variables = {}
            for k, v in loop_values.items():
                key = "{}.{}".format(loop_key, k)
                tmp_variables.update({key: v})
            variables_in_loop.append(tmp_variables)
        else:
            raise ValueError("loop_values of type {} is not supported yet".format(type(loop_values).__name__))

    resolved_opts_in_loop = []
    mutable_vars_per_mo = {}
    for variables in variables_in_loop:
        resolved_opts = None
        if isinstance(taskcall.spec.module_options, dict):
            resolved_opts = {}
            for (
                module_opt_key,
                module_opt_val,
            ) in taskcall.spec.module_options.items():
                if not isinstance(module_opt_val, str):
                    resolved_opts[module_opt_key] = module_opt_val
                    continue
                if not variable_block_re.search(module_opt_val):
                    resolved_opts[module_opt_key] = module_opt_val
                    continue
                # if variables are used in the module option value string
                var_names = extract_variable_names(module_opt_val)
                resolved_opt_val = module_opt_val
                for var_name_dict in var_names:
                    original_block = var_name_dict.get("original", "")
                    var_name = var_name_dict.get("name", "")
                    default_var_name = var_name_dict.get("default", "")
                    resolved_var_val = variables.get(var_name, None)
                    if resolved_var_val is not None:
                        loop_var_type = variables.get("__v_type__", VariableType.Unknown)
                        loop_var_name = variables.get("__v_name__", "")
                        if loop_var_type not in immutable_var_types:
                            if module_opt_key not in mutable_vars_per_mo:
                                mutable_vars_per_mo[module_opt_key] = []
                            mutable_vars_per_mo[module_opt_key].append(loop_var_name)
                    if resolved_var_val is None:
                        resolved_var_val, v_type, resolve_history = context.resolve_variable(var_name)
                        used_variables[var_name] = resolve_history
                        if resolved_var_val is not None:
                            new_var = {
                                "key": var_name,
                                "value": resolved_var_val,
                                "type": v_type,
                            }
                            if not resolved_vars_contains(resolved_vars, new_var):
                                resolved_vars.append(new_var)
                            if v_type not in immutable_var_types:
                                if module_opt_key not in mutable_vars_per_mo:
                                    mutable_vars_per_mo[module_opt_key] = []
                                mutable_vars_per_mo[module_opt_key].append(var_name)
                    if resolved_var_val is None and default_var_name != "":
                        resolved_var_val, v_type, resolve_history = context.resolve_variable(default_var_name)
                        used_variables[default_var_name] = resolve_history
                        if resolved_var_val is not None:
                            new_var = {
                                "key": default_var_name,
                                "value": resolved_var_val,
                                "type": v_type,
                            }
                            if not resolved_vars_contains(resolved_vars, new_var):
                                resolved_vars.append(new_var)
                            if v_type not in immutable_var_types:
                                if module_opt_key not in mutable_vars_per_mo:
                                    mutable_vars_per_mo[module_opt_key] = []
                                mutable_vars_per_mo[module_opt_key].append(var_name)
                    if resolved_var_val is None:
                        new_var = {
                            "key": var_name,
                            "value": None,
                            "type": v_type,
                        }
                        if not resolved_vars_contains(resolved_vars, new_var):
                            resolved_vars.append(new_var)
                        continue
                    if resolved_opt_val == original_block:
                        resolved_opt_val = resolved_var_val
                        break
                    resolved_opt_val = resolved_opt_val.replace(original_block, str(resolved_var_val))
                resolved_opts[module_opt_key] = resolved_opt_val
        elif isinstance(taskcall.spec.module_options, str):
            resolved_opt_val = taskcall.spec.module_options
            if variable_block_re.search(resolved_opt_val):
                var_names = extract_variable_names(taskcall.spec.module_options)
                for var_name_dict in var_names:
                    original_block = var_name_dict.get("original", "")
                    var_name = var_name_dict.get("name", "")
                    default_var_name = var_name_dict.get("default", "")
                    resolved_var_val = variables.get(var_name, None)
                    if resolved_var_val is not None:
                        loop_var_type = variables.get("__v_type__", VariableType.Unknown)
                        loop_var_name = variables.get("__v_name__", "")
                        if loop_var_type not in immutable_var_types:
                            if "" not in mutable_vars_per_mo:
                                mutable_vars_per_mo[""] = []
                            mutable_vars_per_mo[""].append(loop_var_name)
                    if resolved_var_val is None:
                        resolved_var_val, v_type, resolve_history = context.resolve_variable(var_name)
                        used_variables[var_name] = resolve_history
                        if resolved_var_val is not None:
                            new_var = {
                                "key": var_name,
                                "value": resolved_var_val,
                                "type": v_type,
                            }
                            if not resolved_vars_contains(resolved_vars, new_var):
                                resolved_vars.append(new_var)
                            if v_type not in immutable_var_types:
                                if "" not in mutable_vars_per_mo:
                                    mutable_vars_per_mo[""] = []
                                mutable_vars_per_mo[""].append(var_name)
                    if resolved_var_val is None and default_var_name != "":
                        resolved_var_val, v_type, resolve_history = context.resolve_variable(default_var_name)
                        used_variables[default_var_name] = resolve_history
                        if resolved_var_val is not None:
                            new_var = {
                                "key": default_var_name,
                                "value": resolved_var_val,
                                "type": v_type,
                            }
                            if not resolved_vars_contains(resolved_vars, new_var):
                                resolved_vars.append(new_var)
                            if v_type not in immutable_var_types:
                                if "" not in mutable_vars_per_mo:
                                    mutable_vars_per_mo[""] = []
                                mutable_vars_per_mo[""].append(var_name)
                    if resolved_var_val is None:
                        new_var = {
                            "key": var_name,
                            "value": None,
                            "type": v_type,
                        }
                        if not resolved_vars_contains(resolved_vars, new_var):
                            resolved_vars.append(new_var)
                        continue
                    if resolved_opt_val == original_block:
                        resolved_opt_val = resolved_var_val
                        break
                    resolved_opt_val = resolved_opt_val.replace(original_block, str(resolved_var_val))
            resolved_opts = resolved_opt_val
        else:
            resolved_opts = taskcall.spec.module_options
        resolved_opts_in_loop.append(resolved_opts)
    return resolved_opts_in_loop, resolved_vars, mutable_vars_per_mo, used_variables


def extract_variable_names(txt):
    if not variable_block_re.search(txt):
        return []
    found_var_blocks = variable_block_re.findall(txt)
    blocks = []
    for b in found_var_blocks:
        parts = b.split("|")
        var_name = ""
        default_var_name = ""
        for i, p in enumerate(parts):
            if i == 0:
                var_name = p.replace("{{", "").replace("}}", "").replace(" ", "")
            else:
                if "default(" in p and ")" in p:
                    default_var = p.replace("}}", "").replace("default(", "").replace(")", "").replace(" ", "")
                    if not default_var.startswith('"') and not default_var.startswith("'") and not re.compile(r"[0-9].*").match(default_var):
                        default_var_name = default_var
        tmp_b = {
            "original": b,
        }
        if var_name == "":
            continue
        tmp_b["name"] = var_name
        if default_var_name != "":
            tmp_b["default"] = default_var_name
        blocks.append(tmp_b)
    return blocks



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

from ansible_risk_insight.models import (
    AnsibleRunContext,
    RunTargetType,
    Rule,
    Severity,
    RuleTag as Tag,
    VariableType,
    ArgumentsType,
)


@dataclass
class VariableValidationRule(Rule):
    rule_id: str = "P004"
    description: str = "Validate variables and set annotations"
    enabled: bool = True
    name: str = "VariableValidation"
    version: str = "v0.0.1"
    severity: Severity = Severity.NONE
    tags: tuple = Tag.QUALITY
    precedence: int = 0

    def match(self, ctx: AnsibleRunContext) -> bool:
        return ctx.current.type == RunTargetType.Task

    def process(self, ctx: AnsibleRunContext):
        task = ctx.current

        undefined_variables = []
        unknown_name_vars = []
        unnecessary_loop = []
        task_arg_keys = []
        if task.args.type == ArgumentsType.DICT:
            task_arg_keys = list(task.args.raw.keys())
        for v_name in task.variable_use:
            v = task.variable_use[v_name]
            if v and v[-1].type == VariableType.Unknown:
                if v_name not in undefined_variables:
                    undefined_variables.append(v_name)
                if v_name not in unknown_name_vars and v_name not in task_arg_keys:
                    unknown_name_vars.append(v_name)
                if v_name not in unnecessary_loop:
                    if v_name.startswith("item."):
                        unnecessary_loop.append({"name": v_name, "suggested": v_name.replace("item.", "")})

        task.set_annotation("variable.undefined_vars", undefined_variables, rule_id=self.rule_id)
        task.set_annotation("variable.unknown_name_vars", unknown_name_vars, rule_id=self.rule_id)
        task.set_annotation("variable.unnecessary_loop_vars", unnecessary_loop, rule_id=self.rule_id)

        return None



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import json
from copy import deepcopy
from dataclasses import dataclass, field
import ansible_risk_insight.logger as logger
from .keyutil import detect_type, key_delimiter, object_delimiter
from .models import (
    ObjectList,
    Playbook,
    Play,
    RoleInPlay,
    Role,
    Task,
    TaskFile,
    ExecutableType,
    LoadType,
    CallObject,
    TaskCall,
    call_obj_from_spec,
)
from .model_loader import load_builtin_modules
from .risk_assessment_model import RAMClient
from .variable_manager import VariableManager


obj_type_dict = {
    "playbook": "playbooks",
    "play": "plays",
    "role": "roles",
    "taskfile": "taskfiles",
    "task": "tasks",
    "module": "modules",
}

module_name_re = re.compile(r"^[a-z0-9_]+\.[a-z0-9_]+\.[a-z0-9_]+$")
role_name_re = re.compile(r"^[a-z0-9_]+\.[a-z0-9_]+$")
role_in_collection_name_re = re.compile(r"^[a-z0-9_]+\.[a-z0-9_]+\.[a-z0-9_]+$")


@dataclass
class TreeNode(object):
    key: str = ""

    # children is a list of TreeNode
    children: list = field(default_factory=list)

    definition: dict = field(default_factory=dict)

    # load a list of (src, dst) as a tree structure
    # which is composed of multiple TreeNode
    @staticmethod
    def load(graph=[]):
        root_key_cands = [pair[1] for pair in graph if pair[0] is None]
        if len(root_key_cands) != 1:
            raise ValueError("tree array must have only one top with src == None, but" " found {}".format(len(root_key_cands)))
        root_key = root_key_cands[0]
        tree = TreeNode()
        tree, _ = tree.recursive_tree_load(root_key, graph)
        tree.key = root_key
        tree.children = tree.children
        tree.definition = tree.definition
        return tree

    # output list of (src, dst) to stdout or file
    def dump(self, path=""):
        src_dst_array = self.to_graph()
        if path == "":
            print(json.dumps(src_dst_array, indent=2))
        else:
            tree_json = json.dumps(src_dst_array)
            with open(path, "w") as file:
                file.write(tree_json)

    def to_str(self):
        src_dst_array = self.to_graph()
        return json.dumps(src_dst_array)

    # return list of (src, dst)
    def to_graph(self):
        return self.recursive_graph_dump(None, self)

    # return list of dst keys
    def to_keys(self):
        return [p[1] for p in self.to_graph()]

    # reutrn list of TreeNodes that are under this TreeNode
    def to_list(self):
        return self.recursive_convert_to_list(self)

    def recursive_convert_to_list(self, node, nodelist=[]):
        current = [pair for pair in nodelist]
        current.append(node)
        for child_node in node.children:
            current = self.recursive_convert_to_list(child_node, current)
        return current

    def recursive_tree_load(self, node_key, src_dst_array, parent_keys=None):
        if parent_keys is None:
            parent_keys = set()
        n = TreeNode(key=node_key)
        if node_key in parent_keys:
            return n, parent_keys
        parent_keys.add(node_key)
        new_parent_keys = parent_keys.copy()
        for (src_key, dst_key) in src_dst_array:
            children_keys = []
            if node_key == src_key:
                children_keys.append(dst_key)
            for c_key in children_keys:
                child_tree, sub_parent_keys = self.recursive_tree_load(c_key, src_dst_array, parent_keys)
                n.children.append(child_tree)
                new_parent_keys = new_parent_keys.union(sub_parent_keys)
        return n, new_parent_keys

    def recursive_graph_dump(self, parent_node, node, src_dst_array=[]):
        current = [pair for pair in src_dst_array]
        src = None if parent_node is None else parent_node.key
        dst = node.key
        current.append((src, dst))
        for child_node in node.children:
            is_included = len([(src, dst) for (src, dst) in current if src == child_node.key]) > 0
            if is_included:
                continue
            current = self.recursive_graph_dump(node, child_node, current)
        return current

    # return a list of (src, dst) which ends with the "end_key"
    # this could return multiple paths
    def path_to_root(self, end_key):
        path_array = self.search_branch_to_key(end_key, self)
        path_array = [nodelist2branch(nodelist) for nodelist in path_array]
        return path_array

    def search_branch_to_key(self, search_key, node, ancestors=[]):
        current = [n for n in ancestors]
        found = []
        if node.key == search_key:
            found = current + [node]
        for child_node in node.children:
            found_in_child = self.search_branch_to_key(search_key, child_node, current + [node])
            found.extend(found_in_child)
        return found

    def copy(self):
        return deepcopy(self)

    @property
    def is_empty(self):
        return self.key == "" and len(self.children) == 0

    @property
    def has_definition(self):
        return len(self.definition) == 0


def nodelist2branch(nodelist):
    if len(nodelist) == 0:
        return TreeNode()
    t = nodelist[0].copy()
    current = t
    for i, n in enumerate(nodelist):
        if i == 0:
            continue
        current.children = [n.copy()]
        current = current.children[0]
    return t


def load_single_definition(defs: dict, key: str):
    obj_list = ObjectList()
    items = defs.get(key, [])
    for item in items:
        obj_list.add(item)
    return obj_list


def load_definitions(defs: dict, types: list):
    def_list = []
    for type_key in types:
        objs_per_type = load_single_definition(defs, type_key)
        def_list.append(objs_per_type)
    return def_list


def load_all_definitions(definitions: dict):
    _definitions = {}
    if "mappings" in definitions:
        _definitions = {"root": definitions}
    else:
        _definitions = definitions
    loaded = {}
    types = ["roles", "taskfiles", "modules", "playbooks", "plays", "tasks"]
    for type_key in types:
        loaded[type_key] = ObjectList()
    for _, definitions_per_artifact in _definitions.items():
        def_list = load_definitions(definitions_per_artifact.get("definitions", {}), types)
        for i, type_key in enumerate(types):
            if type_key not in loaded:
                loaded[type_key] = def_list[i]
            else:
                loaded[type_key].merge(def_list[i])
    return loaded


def make_dicts(root_definitions, ext_definitions):
    definitions = {
        "roles": ObjectList(),
        "modules": ObjectList(),
        "taskfiles": ObjectList(),
        "playbooks": ObjectList(),
    }
    for type_key in definitions:
        definitions[type_key].merge(root_definitions.get(type_key, ObjectList()))
        definitions[type_key].merge(ext_definitions.get(type_key, ObjectList()))
    dicts = {k: {} for k in definitions}
    for type_key, obj_list in definitions.items():
        for obj in obj_list.items:
            obj_dict_key = obj.fqcn if hasattr(obj, "fqcn") else obj.key
            if type_key not in dicts:
                dicts[type_key] = {}
            dicts[type_key][obj_dict_key] = obj
    return dicts


def resolve(obj, dicts):
    failed = False
    if isinstance(obj, Task):
        task = obj
        if task.executable != "":
            if task.executable_type == ExecutableType.MODULE_TYPE:
                task.resolved_name = resolve_module(task.executable, dicts.get("modules", {}))
            elif task.executable_type == ExecutableType.ROLE_TYPE:
                task.resolved_name = resolve_role(
                    task.executable,
                    dicts.get("roles", {}),
                    task.collection,
                    task.collections_in_play,
                )
            elif task.executable_type == ExecutableType.TASKFILE_TYPE:
                task.resolved_name = resolve_taskfile(task.executable, dicts.get("taskfiles", {}), task.key)
            if task.resolved_name == "":
                failed = True
    elif isinstance(obj, Play):
        for i in range(len(obj.roles)):
            roleinplay = obj.roles[i]
            if not isinstance(roleinplay, RoleInPlay):
                continue
            roleinplay.resolved_name = resolve_role(
                roleinplay.name,
                dicts.get("roles", {}),
                roleinplay.collection,
                roleinplay.collections_in_play,
            )
            obj.roles[i] = roleinplay
            if roleinplay.resolved_name == "":
                failed = True
    return obj, failed


def resolve_module(module_name, module_dict={}):
    module_key = ""
    found_module = module_dict.get(module_name, None)
    if found_module is not None:
        module_key = found_module.key
    if module_key == "":
        for k in module_dict:
            suffix = ".{}".format(module_name)
            if k.endswith(suffix):
                module_key = module_dict[k].key
                break
    return module_key


def resolve_role(role_name, role_dict={}, my_collection_name="", collections_in_play=[]):
    role_key = ""
    if "." not in role_name and len(collections_in_play) > 0:
        for coll in collections_in_play:
            role_name_cand = "{}.{}".format(coll, role_name)
            found_role = role_dict.get(role_name_cand, None)
            if found_role is not None:
                role_key = found_role.key
    else:
        if "." not in role_name and my_collection_name != "":
            role_name_cand = "{}.{}".format(my_collection_name, role_name)
            found_role = role_dict.get(role_name_cand, None)
            if found_role is not None:
                role_key = found_role.key
        if role_key == "":
            found_role = role_dict.get(role_name, None)
            if found_role is not None:
                role_key = found_role.key
            else:
                for k in role_dict:
                    suffix = ".{}".format(role_name)
                    if k.endswith(suffix):
                        role_key = role_dict[k].key
                        break
    return role_key


def resolve_taskfile(taskfile_ref, taskfile_dict={}, task_key=""):
    type_prefix = "task "
    parts = task_key[len(type_prefix) :].split(object_delimiter)
    parent_key = ""
    task_defined_path = ""
    for p in parts[::-1]:
        if p.startswith("playbook" + key_delimiter) or p.startswith("taskfile" + key_delimiter):
            task_defined_path = p.split(key_delimiter)[1]
            parent_key = task_key[len(type_prefix) :].split(p)[0]
            break

    # include/import tasks can have a path like "roles/xxxx/tasks/yyyy.yml"
    # then try to find roles directory
    if taskfile_ref.startswith("roles/"):
        if "roles/" in task_defined_path:
            roles_parent_dir = task_defined_path.split("roles/")[0]
            fpath = os.path.join(roles_parent_dir, taskfile_ref)
            fpath = os.path.normpath(fpath)
            taskfile_key = "taskfile {}taskfile{}{}".format(parent_key, key_delimiter, fpath)
            found_tf = taskfile_dict.get(taskfile_key, None)
            if found_tf is not None:
                return found_tf.key

    task_dir = os.path.dirname(task_defined_path)
    fpath = os.path.join(task_dir, taskfile_ref)
    # need to normalize path here because taskfile_ref can be
    # something like "../some_taskfile.yml".
    # it should be "tasks/some_taskfile.yml"
    fpath = os.path.normpath(fpath)
    taskfile_key = "taskfile {}taskfile{}{}".format(parent_key, key_delimiter, fpath)
    found_tf = taskfile_dict.get(taskfile_key, None)
    if found_tf is not None:
        return found_tf.key

    return ""


def resolve_playbook(playbook_ref, playbook_dict={}, play_key=""):
    type_prefix = "play "
    parts = play_key[len(type_prefix) :].split(object_delimiter)
    parent_key = ""
    play_defined_path = ""
    for p in parts[::-1]:
        if p.startswith("playbook" + key_delimiter):
            play_defined_path = p.split(key_delimiter)[1]
            parent_key = play_key[len(type_prefix) :].split(p)[0]
            break

    play_dir = os.path.dirname(play_defined_path)
    fpath = os.path.join(play_dir, playbook_ref)
    # need to normalize path here because playbook_ref can be
    # something like "../some_playbook.yml"
    fpath = os.path.normpath(fpath)
    playbook_key = "playbook {}playbook{}{}".format(parent_key, key_delimiter, fpath)
    found_playbook = playbook_dict.get(playbook_key, None)
    if found_playbook is not None:
        return found_playbook.key
    return ""


def init_builtin_modules():
    builtin_module_dict = load_builtin_modules()
    modules = list(builtin_module_dict.values())
    return modules


class TreeLoader(object):
    def __init__(self, root_definitions, ext_definitions, ram_client=None, target_playbook_path=None, target_taskfile_path=None):
        self.ram_client: RAMClient = ram_client

        # use mappings just to get tree tops (playbook/role)
        # we don't load any files by this mappings here
        self.load_and_mapping = root_definitions.get("mappings", None)
        self.playbook_mappings = self.load_and_mapping.playbooks
        self.role_mappings = self.load_and_mapping.roles
        self.taskfile_mappings = []

        if target_playbook_path:
            self.playbook_mappings = [p for p in self.playbook_mappings if p[0] == target_playbook_path]
            self.role_mappings = []

        if target_taskfile_path:
            self.taskfile_mappings = self.load_and_mapping.taskfiles
            self.taskfile_mappings = [tf for tf in self.taskfile_mappings if tf[0] == target_taskfile_path]
            self.playbook_mappings = []
            self.role_mappings = []

        # TODO: dependency check, especially for
        # collection dependencies for role

        self.org_root_definitions = root_definitions
        self.org_ext_definitions = ext_definitions

        self.root_definitions = load_all_definitions(root_definitions)
        self.ext_definitions = load_all_definitions(ext_definitions)
        self.add_builtin_modules()

        self.dicts = make_dicts(self.root_definitions, self.ext_definitions)

        self.var_manager: VariableManager = VariableManager()

        self.target_playbook_path = target_playbook_path

        self.module_resolve_cache = {}
        self.role_resolve_cache = {}
        self.taskfile_resolve_cache = {}

        self.resolved_module_from_ram = {}
        self.resolved_role_from_ram = {}
        self.resolved_taskfile_from_ram = {}

        self.extra_requirements = []
        self.extra_requirement_obj_set = set()

        self.trees = []

        self.resolve_failures = {
            "module": {},
            "taskfile": {},
            "role": {},
        }
        return

    def run(self):
        additional_objects = ObjectList()
        if self.load_and_mapping.target_type == LoadType.PROJECT:
            p_defs = self.org_root_definitions.get("definitions", {}).get("projects", [])
            if len(p_defs) > 0:
                additional_objects.add(p_defs[0])
        for i, mapping in enumerate(self.playbook_mappings):
            logger.debug("[{}/{}] {}".format(i + 1, len(self.playbook_mappings), mapping[1]))
            playbook_key = mapping[1]
            tree_objects = self._recursive_get_calls(playbook_key)
            self.trees.append(tree_objects)
        for i, mapping in enumerate(self.role_mappings):
            logger.debug("[{}/{}] {}".format(i + 1, len(self.role_mappings), mapping[1]))
            role_key = mapping[1]
            tree_objects = self._recursive_get_calls(role_key)
            self.trees.append(tree_objects)
        for i, mapping in enumerate(self.taskfile_mappings):
            logger.debug("[{}/{}] {}".format(i + 1, len(self.taskfile_mappings), mapping[1]))
            taskfile_key = mapping[1]
            tree_objects = self._recursive_get_calls(taskfile_key)
            self.trees.append(tree_objects)
        return self.trees, additional_objects

    def _recursive_get_calls(self, key, caller=None, index=0):
        obj_list = ObjectList()
        obj = self.get_object(key)
        if obj is None:
            return obj_list
        call_obj = call_obj_from_spec(spec=obj, caller=caller, index=index)
        if call_obj is not None:
            obj_list.add(call_obj, update_dict=False)
        children_keys, from_ram = self._get_children_keys(obj)
        for i, c_key in enumerate(children_keys):
            child_objects = self._recursive_get_calls(
                c_key,
                call_obj,
                i,
            )
            if isinstance(call_obj, TaskCall):
                taskcall = call_obj
                if len(child_objects.items) > 0:
                    c_obj = child_objects.items[0]
                    if taskcall.spec.executable_type == ExecutableType.MODULE_TYPE:
                        taskcall.module = c_obj.spec
                        if c_key in from_ram:
                            req_info = from_ram[c_key]
                            taskcall.spec.possible_candidates = [(c_obj.spec.fqcn, req_info)]
                        else:
                            taskcall.spec.resolved_name = c_obj.spec.fqcn
                    elif taskcall.spec.executable_type == ExecutableType.ROLE_TYPE:
                        if c_key in from_ram:
                            req_info = from_ram[c_key]
                            taskcall.spec.possible_candidates = [(c_obj.spec.fqcn, req_info)]
                        else:
                            taskcall.spec.resolved_name = c_obj.spec.fqcn
                    elif taskcall.spec.executable_type == ExecutableType.TASKFILE_TYPE:
                        if c_key in from_ram:
                            req_info = from_ram[c_key]
                            taskcall.spec.possible_candidates = [(c_obj.spec.key, req_info)]
                        else:
                            taskcall.spec.resolved_name = c_obj.spec.key
            for c_obj in child_objects.items:
                obj_list.add(c_obj)
        return obj_list

    def _recursive_make_graph(self, key, graph, _objects, caller=None):
        current_graph = [g for g in graph]
        # if this key is already in the graph src, no need to trace children
        key_in_graph_src = [g for g in current_graph if g[0] == key]
        if len(key_in_graph_src) > 0:
            return current_graph
        # otherwise, trace children
        obj = self.get_object(key)
        if obj is None:
            return current_graph
        call_obj = call_obj_from_spec(spec=obj, caller=caller)
        if call_obj is not None:
            caller_key = None if caller is None else caller.key
            my_key = call_obj.key
            current_graph.append([caller_key, my_key])
            _objects.add(call_obj, update_dict=False)
        children_keys = self._get_children_keys(obj)
        for c_key in children_keys:
            updated_graph = self._recursive_make_graph(
                c_key,
                current_graph,
                _objects,
                call_obj,
            )
            current_graph = updated_graph
        return current_graph

    # get definition object from root/ext definitions
    def get_object(self, obj_key, search_ram=False):
        obj_type = detect_type(obj_key)
        if obj_type == "":
            raise ValueError('failed to detect object type from key "{}"'.format(obj_key))
        type_key = obj_type_dict[obj_type]
        root_definitions = self.root_definitions.get(type_key, ObjectList())
        obj = root_definitions.find_by_key(obj_key)
        if obj is not None:
            return obj
        ext_definitions = self.ext_definitions.get(type_key, ObjectList())
        obj = ext_definitions.find_by_key(obj_key)
        if obj is not None:
            return obj

        if search_ram and self.ram_client:
            matched_obj = self.ram_client.get_object_by_key(obj_key)
            obj = matched_obj.get("object", None)
            if obj is not None:
                return obj

        return None

    def add_builtin_modules(self):
        builtin_module_dict = {}
        if self.ram_client and self.ram_client.builtin_modules_cache:
            builtin_module_dict = self.ram_client.builtin_modules_cache
        else:
            builtin_module_dict = load_builtin_modules()
            if self.ram_client:
                self.ram_client.builtin_modules_cache = builtin_module_dict
        builtin_modules = list(builtin_module_dict.values())
        obj_list = ObjectList(items=builtin_modules)
        self.ext_definitions["modules"].merge(obj_list)

    def _get_children_keys(self, obj):
        if isinstance(obj, CallObject):
            return self._get_children_keys(obj.spec)
        children_keys = []
        from_ram = {}
        if isinstance(obj, Playbook):
            children_keys = obj.plays
        elif isinstance(obj, Play):
            if obj.import_playbook != "":
                resolved_playbook_key = resolve_playbook(obj.import_playbook, self.dicts["playbooks"], obj.key)
                if resolved_playbook_key != "":
                    children_keys.append(resolved_playbook_key)
            children_keys.extend(obj.pre_tasks)
            children_keys.extend(obj.tasks)
            for rip in obj.roles:

                if rip.name in self.role_resolve_cache:
                    resolved_role_key = self.role_resolve_cache[rip.name]
                else:
                    resolved_role_key = resolve_role(
                        rip.name,
                        self.dicts["roles"],
                        obj.collection,
                        obj.collections_in_play,
                    )
                    if resolved_role_key != "":
                        self.role_resolve_cache[rip.name] = resolved_role_key

                if resolved_role_key == "" and self.ram_client is not None:
                    if rip.name in self.resolved_role_from_ram:
                        resolved_role_key, req_info = self.resolved_role_from_ram[rip.name]
                        from_ram[resolved_role_key] = req_info
                    else:
                        matched_roles = self.ram_client.search_role(rip.name)
                        if len(matched_roles) > 0:
                            resolved_role_key = matched_roles[0]["object"].key
                            self.ext_definitions["roles"].add(matched_roles[0]["object"])
                            for offspr_obj in matched_roles[0].get("offspring_objects", []):
                                type_str = offspr_obj["type"] + "s"
                                self.ext_definitions[type_str].add(offspr_obj["object"])
                            if matched_roles[0]["object"].key not in self.extra_requirement_obj_set:
                                self.extra_requirements.append(
                                    {
                                        "type": "role",
                                        "name": matched_roles[0]["object"].fqcn,
                                        "defined_in": matched_roles[0]["defined_in"],
                                        "used_in": obj.defined_in,
                                    }
                                )
                                self.extra_requirement_obj_set.add(matched_roles[0]["object"].key)
                            for offspr_obj in matched_roles[0].get("offspring_objects", []):
                                if hasattr(offspr_obj["object"], "builtin") and offspr_obj["object"].builtin:
                                    continue
                                if offspr_obj["object"].key not in self.extra_requirement_obj_set:
                                    self.extra_requirements.append(
                                        {
                                            "type": offspr_obj["type"],
                                            "name": offspr_obj["name"],
                                            "defined_in": offspr_obj["defined_in"],
                                            "used_in": offspr_obj["used_in"],
                                        }
                                    )
                                    self.extra_requirement_obj_set.add(offspr_obj["object"].key)
                            self.resolved_role_from_ram[rip.name] = (resolved_role_key, matched_roles[0]["defined_in"])
                            from_ram[resolved_role_key] = matched_roles[0]["defined_in"]

                if resolved_role_key != "":
                    children_keys.append(resolved_role_key)
            children_keys.extend(obj.post_tasks)
        elif isinstance(obj, Role):
            main_taskfile_key = [tf for tf in obj.taskfiles if tf.split(key_delimiter)[-1].split("/")[-1] in ["main.yml", "main.yaml"]]
            children_keys.extend(main_taskfile_key)
        elif isinstance(obj, TaskFile):
            children_keys = obj.tasks
        elif isinstance(obj, Task):
            executable_type = obj.executable_type
            resolved_key = ""
            if obj.executable == "":
                return [], {}
            target_name = obj.executable
            if executable_type == ExecutableType.MODULE_TYPE:
                if target_name in self.module_resolve_cache:
                    resolved_key = self.module_resolve_cache[target_name]
                else:
                    resolved_key = resolve_module(target_name, self.dicts["modules"])
                    if resolved_key != "":
                        self.module_resolve_cache[target_name] = resolved_key
                if resolved_key == "" and self.ram_client is not None:
                    if target_name in self.resolved_module_from_ram:
                        resolved_key, req_info = self.resolved_module_from_ram[target_name]
                        from_ram[resolved_key] = req_info
                    else:
                        matched_modules = self.ram_client.search_module(target_name)
                        if len(matched_modules) > 0:
                            resolved_key = matched_modules[0]["object"].key
                            self.ext_definitions["modules"].add(matched_modules[0]["object"])
                            if matched_modules[0]["object"].key not in self.extra_requirement_obj_set:
                                if not matched_modules[0]["object"].builtin:
                                    self.extra_requirements.append(
                                        {
                                            "type": "module",
                                            "name": matched_modules[0]["object"].fqcn,
                                            "defined_in": matched_modules[0]["defined_in"],
                                            "used_in": obj.defined_in,
                                        }
                                    )
                                    self.extra_requirement_obj_set.add(matched_modules[0]["object"].key)
                            self.resolved_module_from_ram[target_name] = (resolved_key, matched_modules[0]["defined_in"])
                            from_ram[resolved_key] = matched_modules[0]["defined_in"]
                if resolved_key == "":
                    if target_name not in self.resolve_failures["module"]:
                        self.resolve_failures["module"][target_name] = 0
                    self.resolve_failures["module"][target_name] += 1
            elif executable_type == ExecutableType.ROLE_TYPE:
                if target_name in self.role_resolve_cache:
                    resolved_key = self.role_resolve_cache[target_name]
                else:
                    resolved_key = resolve_role(
                        target_name,
                        self.dicts["roles"],
                        obj.collection,
                        obj.collections_in_play,
                    )
                    if resolved_key != "":
                        self.role_resolve_cache[target_name] = resolved_key
                if resolved_key == "" and self.ram_client is not None:
                    if target_name in self.resolved_role_from_ram:
                        resolved_key, req_info = self.resolved_role_from_ram[target_name]
                        from_ram[resolved_key] = req_info
                    else:
                        matched_roles = self.ram_client.search_role(target_name)
                        if len(matched_roles) > 0:
                            resolved_key = matched_roles[0]["object"].key
                            self.ext_definitions["roles"].add(matched_roles[0]["object"])
                            if matched_roles[0]["object"].key not in self.extra_requirement_obj_set:
                                self.extra_requirements.append(
                                    {
                                        "type": "role",
                                        "name": matched_roles[0]["object"].fqcn,
                                        "defined_in": matched_roles[0]["defined_in"],
                                        "used_in": obj.defined_in,
                                    }
                                )
                                self.extra_requirement_obj_set.add(matched_roles[0]["object"].key)
                            for offspr_obj in matched_roles[0].get("offspring_objects", []):
                                if hasattr(offspr_obj["object"], "builtin") and offspr_obj["object"].builtin:
                                    continue
                                if offspr_obj["object"].key not in self.extra_requirement_obj_set:
                                    self.extra_requirements.append(
                                        {
                                            "type": offspr_obj["type"],
                                            "name": offspr_obj["name"],
                                            "defined_in": offspr_obj["defined_in"],
                                            "used_in": offspr_obj["used_in"],
                                        }
                                    )
                                    self.extra_requirement_obj_set.add(offspr_obj["object"].key)
                            self.resolved_role_from_ram[target_name] = (resolved_key, matched_roles[0]["defined_in"])
                            from_ram[resolved_key] = matched_roles[0]["defined_in"]
                if resolved_key == "":
                    if target_name not in self.resolve_failures["role"]:
                        self.resolve_failures["role"][target_name] = 0
                    self.resolve_failures["role"][target_name] += 1
            elif executable_type == ExecutableType.TASKFILE_TYPE:
                if is_templated(target_name):
                    target_name = render_template(target_name)
                if target_name in self.taskfile_resolve_cache:
                    resolved_key = self.taskfile_resolve_cache[target_name]
                else:
                    resolved_key = resolve_taskfile(
                        target_name,
                        self.dicts["taskfiles"],
                        obj.key,
                    )
                    if resolved_key != "":
                        self.taskfile_resolve_cache[target_name] = resolved_key
                if resolved_key == "" and self.ram_client is not None:
                    if obj.executable in self.resolved_role_from_ram:
                        resolved_key, req_info = self.resolved_role_from_ram[target_name]
                        from_ram[resolved_key] = req_info
                    else:
                        matched_taskfiles = self.ram_client.search_taskfile(target_name, include_task_path=obj.defined_in)
                        if len(matched_taskfiles) > 0:
                            resolved_key = matched_taskfiles[0]["object"].key
                            self.ext_definitions["taskfiles"].add(matched_taskfiles[0]["object"], update_dict=False)
                            if matched_taskfiles[0]["object"].key not in self.extra_requirement_obj_set:
                                self.extra_requirements.append(
                                    {
                                        "type": "taskfile",
                                        "name": matched_taskfiles[0]["object"].key,
                                        "defined_in": matched_taskfiles[0]["defined_in"],
                                        "used_in": obj.defined_in,
                                    }
                                )
                                self.extra_requirement_obj_set.add(matched_taskfiles[0]["object"].key)
                            for offspr_obj in matched_taskfiles[0].get("offspring_objects", []):
                                if hasattr(offspr_obj["object"], "builtin") and offspr_obj["object"].builtin:
                                    continue
                                if offspr_obj["object"].key not in self.extra_requirement_obj_set:
                                    self.extra_requirements.append(
                                        {
                                            "type": offspr_obj["type"],
                                            "name": offspr_obj["name"],
                                            "defined_in": offspr_obj["defined_in"],
                                            "used_in": offspr_obj["used_in"],
                                        }
                                    )
                                    self.extra_requirement_obj_set.add(offspr_obj["object"].key)
                            self.resolved_taskfile_from_ram[target_name] = (resolved_key, matched_taskfiles[0]["defined_in"])
                            from_ram[resolved_key] = matched_taskfiles[0]["defined_in"]
                if resolved_key == "":
                    if target_name not in self.resolve_failures["taskfile"]:
                        self.resolve_failures["taskfile"][target_name] = 0
                    self.resolve_failures["taskfile"][target_name] += 1

            if resolved_key != "":
                children_keys.append(resolved_key)
        return children_keys, from_ram

    def node_objects(self, tree):
        loaded = {}
        obj_list = ObjectList()
        for k in tree.to_keys():
            if k in loaded:
                obj_list.add(loaded[k])
                continue
            obj = self.get_object(k)
            if obj is None:
                logger.warning("object not found for the key {}".format(k))
                continue
            obj_list.add(obj)
            loaded[k] = obj
        return obj_list


def is_templated(txt):
    return "{{" in txt


# TODO: need to use variable manager
def render_template(txt, variable_manager=None):
    regex = r'[\'"]([^\'"]+\.ya?ml)[\'"]'
    matched = re.search(regex, txt)
    if matched:
        return matched.group(1)
    if "{{ ansible_facts.os_family }}.yml" in txt:
        return "Debian.yml"
    if "{{ gcloud_install_type }}/main.yml" in txt:
        return "package/main.yml"
    if "{{ ansible_os_family | lower }}.yml" in txt:
        return "debian.yml"
    return txt


def dump_node_objects(obj_list, path=""):
    if path == "":
        lines = obj_list.dump()
        for line in lines:
            obj_dict = json.loads(line)
            print(json.dumps(obj_dict, indent=2))
    else:
        obj_list.dump(fpath=path)


def key_to_file_name(prefix, key):
    return prefix + "___" + key.translate(str.maketrans({" ": "___", "/": "---", ".": "_dot_"})) + ".json"



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from ansible_risk_insight.models import TaskCall, RiskAnnotation
from ansible_risk_insight.utils import load_classes_in_dir
from ansible_risk_insight.annotators.annotator_base import Annotator, AnnotatorResult
from ansible_risk_insight.annotators.module_annotator_base import ModuleAnnotator, ModuleAnnotatorResult


class RiskAnnotator(Annotator):
    type: str = RiskAnnotation.type
    name: str = ""
    enabled: bool = False

    module_annotator_cache: dict = {}

    def match(self, task: TaskCall) -> bool:
        raise ValueError("this is a base class method")

    def run(self, task: TaskCall):
        raise ValueError("this is a base class method")

    def load_module_annotators(self, dir_path: str):
        if dir_path in self.module_annotator_cache:
            return self.module_annotator_cache[dir_path]

        annotator_classes, _ = load_classes_in_dir(dir_path, ModuleAnnotator, __file__)
        module_annotators = []
        for a_c in annotator_classes:
            annotator = a_c(context=self.context)
            module_annotators.append(annotator)
        if module_annotators:
            self.module_annotator_cache[dir_path] = module_annotators
        return module_annotators

    def run_module_annotators(self, dir_path: str, task: TaskCall) -> ModuleAnnotatorResult:
        if not dir_path:
            return []

        resolved_name = task.spec.resolved_name
        module_annotators = self.load_module_annotators(dir_path)

        # TODO: need to consider annotator precedence

        annotations = []

        for annotator in module_annotators:
            if not isinstance(annotator, ModuleAnnotator):
                continue
            if not annotator.fqcn:
                continue
            if resolved_name != annotator.fqcn:
                continue

            result = annotator.run(task)
            if not result:
                continue

            if result.annotations:
                annotations.extend(result.annotations)
        if annotations:
            return ModuleAnnotatorResult(annotations=annotations)
        return None


@dataclass
class RiskAnnotatorResult(AnnotatorResult):
    pass



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

from ansible_risk_insight.models import (
    AnsibleRunContext,
    RunTargetType,
    DefaultRiskType as RiskType,
    AnnotationCondition,
    Rule,
    Severity,
    RuleTag as Tag,
    RuleResult,
)


@dataclass
class ListAllInboundSrcRule(Rule):
    rule_id: str = "R401"
    description: str = "List all inbound sources"
    enabled: bool = True
    name: str = "ListAllInboundSrcRule"
    version: str = "v0.0.1"
    severity: Severity = Severity.VERY_LOW
    tags: tuple = Tag.DEBUG

    def match(self, ctx: AnsibleRunContext) -> bool:
        return ctx.current.type == RunTargetType.Task

    def process(self, ctx: AnsibleRunContext):
        task = ctx.current

        ac = AnnotationCondition().risk_type(RiskType.INBOUND)
        verdict = False
        detail = {}
        src_list = []
        if ctx.is_end(task):
            tasks = ctx.search(ac)
            for t in tasks:
                anno = t.get_annotation_by_condition(ac)
                if anno:
                    src_list.append(anno.src.value)
            if len(src_list) > 0:
                verdict = True
                detail["inbound_src"] = src_list

        return RuleResult(verdict=verdict, detail=detail, file=task.file_info(), rule=self.get_metadata())



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import yaml
import json
import subprocess
import ansible_risk_insight.logger as logger
from .safe_glob import safe_glob

from .models import (
    LoadType,
)


collection_manifest_json = "MANIFEST.json"
role_meta_main_yml = "meta/main.yml"
role_meta_main_yaml = "meta/main.yaml"
requirements_yml = "requirements.yml"
galaxy_yml = "galaxy.yml"
GALAXY_yml = "GALAXY.yml"


def find_dependency(type, target, dependency_dir):
    dependencies = {"dependencies": {}, "type": "", "file": ""}
    logger.debug("search dependency")
    if dependency_dir:
        requirements, paths, metadata = load_existing_dependency_dir(dependency_dir)
        dependencies["dependencies"] = requirements
        dependencies["paths"] = paths
        dependencies["metadata"] = metadata
        dependencies["type"] = type
        dependencies["file"] = None
    else:
        if type == LoadType.PROJECT:
            logger.debug("search project dependency")
            requirements, reqyml = find_project_dependency(target)
            dependencies["dependencies"] = requirements
            dependencies["type"] = LoadType.PROJECT
            dependencies["file"] = reqyml
        elif type == LoadType.ROLE:
            logger.debug("search role dependency")
            requirements, mainyml = find_role_dependency(target)
            dependencies["dependencies"] = requirements
            dependencies["type"] = LoadType.ROLE
            dependencies["file"] = mainyml
        elif type == LoadType.COLLECTION:
            logger.debug("search collection dependency")
            requirements, manifestjson = find_collection_dependency(target)
            dependencies["dependencies"] = requirements
            dependencies["type"] = LoadType.COLLECTION
            dependencies["file"] = manifestjson

    return dependencies


def find_role_dependency(target):
    requirements = {}
    if not os.path.exists(target):
        raise ValueError("Invalid target dir: {}".format(target))
    role_meta_files = safe_glob(
        [
            os.path.join(target, "**", role_meta_main_yml),
            os.path.join(target, "**", role_meta_main_yaml),
        ],
        recursive=True,
    )
    main_yaml = ""
    if len(role_meta_files) > 0:
        for rf in role_meta_files:
            if os.path.exists(rf):
                main_yaml = rf
                with open(rf, "r") as file:
                    try:
                        metadata = yaml.safe_load(file)
                    except Exception as e:
                        logger.debug("failed to load this yaml file to read metadata; {}".format(e.args[0]))

                    if metadata is not None and isinstance(metadata, dict):
                        requirements["roles"] = metadata.get("dependencies", [])
                        requirements["collections"] = metadata.get("collections", [])
    return requirements, main_yaml


def find_collection_dependency(target):
    requirements = {}
    # collection dir installed by ansible-galaxy command
    manifest_json_files = safe_glob(os.path.join(target, "**", collection_manifest_json), recursive=True)
    logger.debug("found meta files {}".format(manifest_json_files))
    manifest_json = ""
    if len(manifest_json_files) > 0:
        for cmf in manifest_json_files:
            if os.path.exists(cmf):
                manifest_json = cmf
                metadata = {}
                with open(cmf, "r") as file:
                    metadata = json.load(file)
                    dependencies = metadata.get("collection_info", {}).get("dependencies", [])
                    requirements["collections"] = format_dependency_info(dependencies)
    else:
        requirements, manifest_json = load_dependency_from_galaxy(target)
    return requirements, manifest_json


def find_project_dependency(target):
    if os.path.exists(target):
        # local dir
        logger.debug("load requirements from dir {}".format(target))
        return load_requirements(target)
    else:
        raise ValueError("Invalid target dir: {}".format(target))


def load_requirements(path):
    requirements = {}
    yaml_path = ""
    # project dir
    requirements_yml_path = os.path.join(path, requirements_yml)
    if os.path.exists(requirements_yml_path):
        yaml_path = requirements_yml_path
        with open(requirements_yml_path, "r") as file:
            try:
                requirements = yaml.safe_load(file)
            except Exception as e:
                logger.debug("failed to load requirements.yml; {}".format(e.args[0]))
    else:
        requirements, yaml_path = load_dependency_from_galaxy(path)
    return requirements, yaml_path


def load_dependency_from_galaxy(path):
    requirements = {}
    yaml_path = ""
    galaxy_yml_files = safe_glob(os.path.join(path, "**", galaxy_yml), recursive=True)
    galaxy_yml_files.extend(safe_glob(os.path.join(path, "**", GALAXY_yml), recursive=True))
    logger.debug("found meta files {}".format(galaxy_yml_files))
    if len(galaxy_yml_files) > 0:
        for g in galaxy_yml_files:
            if os.path.exists(g):
                yaml_path = g
                metadata = {}
                with open(g, "r") as file:
                    metadata = yaml.safe_load(file)
                    dependencies = metadata.get("dependencies", {})
                    requirements["collections"] = format_dependency_info(dependencies)
    return requirements, yaml_path


def load_existing_dependency_dir(dependency_dir):
    # role_meta_files = safe_glob(
    #     [
    #         os.path.join(dependency_dir, "**", role_meta_main_yml),
    #         os.path.join(dependency_dir, "**", role_meta_main_yaml),
    #     ],
    #     recursive=True,
    # )
    collection_meta_files = safe_glob(os.path.join(dependency_dir, "**", collection_manifest_json), recursive=True)
    requirements = {
        "roles": [],
        "collections": [],
    }
    paths = {
        "roles": {},
        "collections": {},
    }
    metadata = {
        "roles": {},
        "collections": {},
    }
    # for r_meta_file in role_meta_files:
    #     role_name = r_meta_file.split("/")[-3]
    #     role_path = "/".join(r_meta_file.split("/")[:-2])
    #     requirements["roles"].append(role_name)
    #     paths["roles"][role_name] = role_path
    for c_meta_file in collection_meta_files:
        if "/tests/" in c_meta_file:
            continue
        parts = c_meta_file.split("/")
        collection_name = f"{parts[-3]}.{parts[-2]}"
        collection_path = "/".join(c_meta_file.split("/")[:-1])
        collection_base_path = "/".join(c_meta_file.split("/")[:-3])
        for dir_name in os.listdir(collection_base_path):
            galaxy_data = None
            if dir_name.startswith(collection_name) and dir_name.endswith(".info"):
                galaxy_yml_path = os.path.join(collection_base_path, dir_name, galaxy_yml)
                try:
                    with open(galaxy_yml_path, "r") as galaxy_yml_file:
                        galaxy_data = yaml.safe_load(galaxy_yml_file)
                except Exception:
                    pass
            if not isinstance(galaxy_data, dict):
                continue
            metadata["collections"][collection_name] = galaxy_data
        requirements["collections"].append(collection_name)
        paths["collections"][collection_name] = collection_path
    return requirements, paths, metadata


def install_github_target(target, output_dir):
    proc = subprocess.run(
        "git clone {} {}".format(target, output_dir),
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )
    install_msg = proc.stdout
    logger.debug("STDOUT: {}".format(install_msg))
    return proc.stdout


def format_dependency_info(dependencies):
    results = []
    for k, v in dependencies.items():
        results.append({"name": k, "version": v})
    return results



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import json
from typing import List
from ansible_risk_insight.annotators.risk_annotator_base import RiskAnnotator
import ansible_risk_insight.logger as logger
from .models import TaskCallsInTree, AnsibleRunContext
from .utils import load_classes_in_dir


def load_annotators(ctx: AnsibleRunContext = None):
    _annotator_classes, _ = load_classes_in_dir("annotators", RiskAnnotator, __file__)
    _annotators = []
    for a in _annotator_classes:
        try:
            _annotator = a(context=ctx)
            _annotators.append(_annotator)
        except Exception:
            raise ValueError(f"failed to load an annotator: {a}")
    return _annotators


def load_taskcalls_in_trees(path: str) -> List[TaskCallsInTree]:
    taskcalls_in_trees = []
    try:
        with open(path, "r") as file:
            for line in file:
                taskcalls_in_tree = TaskCallsInTree.from_json(line)
                taskcalls_in_trees.append(taskcalls_in_tree)
    except Exception as e:
        raise ValueError("failed to load the json file {} {}".format(path, e))
    return taskcalls_in_trees


def analyze(contexts: List[AnsibleRunContext]):
    num = len(contexts)
    for i, ctx in enumerate(contexts):
        if not isinstance(ctx, AnsibleRunContext):
            continue
        for j, t in enumerate(ctx.tasks):
            annotator = None
            _annotators = load_annotators(ctx)
            for ax in _annotators:
                if not ax.enabled:
                    continue
                if ax.match(task=t):
                    annotator = ax
                    break
            if annotator is None:
                continue
            result = annotator.run(task=t)
            if not result:
                continue
            if result.annotations:
                t.annotations.extend(result.annotations)
        logger.debug("analyze() {}/{} done".format(i + 1, num))
    return contexts


def main():
    parser = argparse.ArgumentParser(
        prog="analyze.py",
        description="analyze tasks",
        epilog="end",
        add_help=True,
    )

    parser.add_argument(
        "-i",
        "--input",
        default="",
        help="path to the input json (taskcalls_in_trees.json)",
    )
    parser.add_argument("-o", "--output", default="", help="path to the output json")

    args = parser.parse_args()

    taskcalls_in_trees = load_taskcalls_in_trees(args.input)
    taskcalls_in_trees = analyze(taskcalls_in_trees)

    if args.output != "":
        lines = [json.dumps(single_tree_data) for single_tree_data in taskcalls_in_trees]
        with open(args.output, mode="wt") as file:
            file.write("\n".join(lines))


if __name__ == "__main__":
    main()



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import json
import os
import re
import yaml
from ansible.module_utils.parsing.convert_bool import boolean

import ansible_risk_insight.logger as logger
from .safe_glob import safe_glob
from .models import (
    ExecutableType,
    Inventory,
    InventoryType,
    LoadType,
    Play,
    Playbook,
    PlaybookFormatError,
    Repository,
    Role,
    Module,
    ModuleArgument,
    RoleInPlay,
    Task,
    TaskFile,
    TaskFormatError,
    Collection,
    BecomeInfo,
    ObjectList,
)
from .finder import (
    find_best_repo_root_path,
    find_collection_name_of_repo,
    find_module_name,
    get_task_blocks,
    search_inventory_files,
    find_module_dirs,
    search_module_files,
    search_taskfiles_for_playbooks,
    module_dir_patterns,
)
from .utils import (
    split_target_playbook_fullpath,
    split_target_taskfile_fullpath,
    get_module_specs_by_ansible_doc,
    get_documentation_in_module_file,
    get_class_by_arg_type,
)
from .awx_utils import could_be_playbook

# collection info direcotry can be something like
#   "brightcomputing.bcm-9.1.11+41615.gitfab9053.info"
collection_info_dir_re = re.compile(r"^[a-z0-9_]+\.[a-z0-9_]+-[0-9]+\.[0-9]+\.[0-9]+.*\.info$")

string_module_options_re = re.compile(r"[a-z0-9]+=(?:[^ ]*{{ [^ ]+ }}[^ ]*|[^ ])+")

loop_task_option_names = [
    "loop",
    "with_list",
    "with_items",
    "with_dict",
    # TODO: support the following
    "with_indexed_items",
    "with_flattened",
    "with_together",
    "with_sequence",
    "with_subelements",
    "with_nested",
    "with_cartesian",
    "with_random_choice",
]


def load_repository(
    path="",
    installed_collections_path="",
    installed_roles_path="",
    my_collection_name="",
    basedir="",
    target_playbook_path="",
    target_taskfile_path="",
    use_ansible_doc=True,
    skip_playbook_format_error=True,
    skip_task_format_error=True,
    load_children=True,
):
    repoObj = Repository()

    repo_path = ""
    if path == "":
        # if path is empty, just load installed collections / roles
        repo_path = ""
    else:
        # otherwise, find the root path by searching playbooks
        try:
            repo_path = find_best_repo_root_path(path)
        except Exception:
            logger.debug('failed to find a root directory for ansible files; use "{}"' " but this may be wrong".format(path))
        if repo_path == "":
            repo_path = path

    if repo_path != "":
        if my_collection_name == "":
            my_collection_name = find_collection_name_of_repo(repo_path)
        if my_collection_name != "":
            repoObj.my_collection_name = my_collection_name

    if basedir == "":
        basedir = path

    logger.debug("start loading the repo {}".format(repo_path))
    logger.debug("start loading playbooks")
    repoObj.playbooks = load_playbooks(repo_path, basedir=basedir, load_children=load_children)
    logger.debug("done ... {} playbooks loaded".format(len(repoObj.playbooks)))
    logger.debug("start loading roles")
    repoObj.roles = load_roles(repo_path, basedir=basedir, use_ansible_doc=use_ansible_doc, load_children=load_children)
    logger.debug("done ... {} roles loaded".format(len(repoObj.roles)))
    logger.debug("start loading modules (that are defined in this repository)")
    repoObj.modules = load_modules(
        repo_path,
        basedir=basedir,
        collection_name=repoObj.my_collection_name,
        use_ansible_doc=use_ansible_doc,
        load_children=load_children,
    )
    logger.debug("done ... {} modules loaded".format(len(repoObj.modules)))
    logger.debug("start loading taskfiles (that are defined for playbooks in this" " repository)")
    repoObj.taskfiles = load_taskfiles(repo_path, basedir=basedir, load_children=load_children)
    logger.debug("done ... {} task files loaded".format(len(repoObj.taskfiles)))
    logger.debug("start loading inventory files")
    repoObj.inventories = load_inventories(repo_path, basedir=basedir)
    logger.debug("done ... {} inventory files loaded".format(len(repoObj.inventories)))
    logger.debug("start loading installed collections")
    repoObj.installed_collections = load_installed_collections(installed_collections_path)

    logger.debug("done ... {} collections loaded".format(len(repoObj.installed_collections)))
    logger.debug("start loading installed roles")
    repoObj.installed_roles = load_installed_roles(installed_roles_path)
    logger.debug("done ... {} roles loaded".format(len(repoObj.installed_roles)))
    repoObj.requirements = load_requirements(path=repo_path)
    name = os.path.basename(path)
    repoObj.name = name
    _path = name
    if os.path.abspath(repo_path).startswith(os.path.abspath(path)):
        relative = os.path.abspath(repo_path)[len(os.path.abspath(path)) :]
        _path = os.path.join(name, relative)
    repoObj.path = _path
    repoObj.installed_collections_path = installed_collections_path
    repoObj.installed_roles_path = installed_roles_path
    repoObj.target_playbook_path = target_playbook_path
    repoObj.target_taskfile_path = target_taskfile_path
    logger.debug("done")

    return repoObj


def load_installed_collections(installed_collections_path):
    search_path = installed_collections_path
    if installed_collections_path == "" or not os.path.exists(search_path):
        return []
    if os.path.exists(os.path.join(search_path, "ansible_collections")):
        search_path = os.path.join(search_path, "ansible_collections")
    dirs = os.listdir(search_path)
    basedir = os.path.dirname(os.path.normpath(installed_collections_path))
    collections = []
    for d in dirs:
        if collection_info_dir_re.match(d):
            continue
        if not os.path.exists(os.path.join(search_path, d)):
            continue
        subdirs = os.listdir(os.path.join(search_path, d))
        for sd in subdirs:
            collection_path = os.path.join(search_path, d, sd)
            try:
                c = load_collection(collection_dir=collection_path, basedir=basedir)
                collections.append(c)
            except Exception:
                logger.exception("error while loading the collection at {}".format(collection_path))
    return collections


def load_inventory(path, basedir=""):
    invObj = Inventory()
    fullpath = ""
    if os.path.exists(path) and path != "" and path != ".":
        fullpath = path
    if os.path.exists(os.path.join(basedir, path)):
        fullpath = os.path.normpath(os.path.join(basedir, path))
    if fullpath == "":
        raise ValueError("file not found")
    defined_in = fullpath
    if basedir != "":
        if defined_in.startswith(basedir):
            defined_in = defined_in[len(basedir) :]
            if defined_in.startswith("/"):
                defined_in = defined_in[1:]
    invObj.defined_in = defined_in
    base_parts = os.path.splitext(os.path.basename(fullpath))
    invObj.name = base_parts[0]
    file_ext = base_parts[1]
    dirname = os.path.dirname(fullpath)
    inventory_type = InventoryType.UNKNOWN_TYPE
    group_name = ""
    host_name = ""
    if dirname.endswith("/group_vars"):
        inventory_type = InventoryType.GROUP_VARS_TYPE
        group_name = base_parts[0]
    elif dirname.endswith("/host_vars"):
        inventory_type = InventoryType.HOST_VARS_TYPE
        host_name = base_parts[0]
    invObj.inventory_type = inventory_type
    invObj.group_name = group_name
    invObj.host_name = host_name
    data = {}
    if file_ext == "":
        # TODO: parse it as INI file
        pass
    elif file_ext == ".yml" or file_ext == ".yaml":
        with open(fullpath, "r") as file:
            try:
                data = yaml.safe_load(file)
            except Exception as e:
                logger.debug("failed to load this yaml file (inventory); {}".format(e.args[0]))
    elif file_ext == ".json":
        with open(fullpath, "r") as file:
            try:
                data = json.load(file)
            except Exception as e:
                logger.debug("failed to load this json file (inventory); {}".format(e.args[0]))
    invObj.variables = data
    return invObj


def load_inventories(path, basedir=""):

    if not os.path.exists(path):
        return []
    inventories = []
    inventory_file_paths = search_inventory_files(path)
    if len(inventory_file_paths) > 0:
        for inventory_path in inventory_file_paths:
            try:
                iv = load_inventory(inventory_path, basedir=basedir)
                inventories.append(iv)
            except Exception:
                logger.exception("error while loading the inventory file at {}".format(inventory_path))
    return inventories


def load_play(
    path,
    index,
    play_block_dict,
    role_name="",
    collection_name="",
    parent_key="",
    parent_local_key="",
    yaml_lines="",
    basedir="",
    skip_task_format_error=True,
):
    pbObj = Play()
    if play_block_dict is None:
        raise ValueError("play block dict is required to load Play")
    if not isinstance(play_block_dict, dict):
        raise PlaybookFormatError("this play block is not loaded as dict; maybe this is not a" " playbook")
    data_block = play_block_dict
    if "hosts" not in data_block and "import_playbook" not in data_block and "include" not in data_block:
        raise PlaybookFormatError('this play block does not have "hosts", "import_playbook" and' ' "include"; maybe this is not a playbook')

    pbObj.index = index
    pbObj.role = role_name
    pbObj.collection = collection_name
    pbObj.set_key(parent_key, parent_local_key)
    play_name = data_block.get("name", "")
    collections_in_play = data_block.get("collections", [])
    pre_tasks = []
    post_tasks = []
    tasks = []
    roles = []
    play_options = {}
    import_module = ""
    import_playbook = ""
    __pre_task_blocks, _ = get_task_blocks(task_dict_list=data_block.get("pre_tasks", []))
    __task_blocks, _ = get_task_blocks(task_dict_list=data_block.get("tasks", []))
    pre_task_num = len(__pre_task_blocks) if __pre_task_blocks else 0
    task_num = len(__task_blocks) if __task_blocks else 0
    for k, v in data_block.items():
        if k == "name":
            pass
        elif k == "collections":
            pass
        elif k == "pre_tasks":
            if not isinstance(v, list):
                continue
            task_blocks, _ = get_task_blocks(task_dict_list=v)
            if task_blocks is None:
                continue
            for i, task_dict in enumerate(task_blocks):
                try:
                    t = load_task(
                        path=path,
                        index=i,
                        task_block_dict=task_dict,
                        role_name=role_name,
                        collection_name=collection_name,
                        collections_in_play=collections_in_play,
                        play_index=index,
                        parent_key=pbObj.key,
                        parent_local_key=pbObj.local_key,
                        yaml_lines=yaml_lines,
                        basedir=basedir,
                    )
                    pre_tasks.append(t)
                except TaskFormatError:
                    if skip_task_format_error:
                        logger.debug("this task is wrong format; skip the task in {}," " index: {}; skip this".format(path, i))
                        continue
                    else:
                        raise TaskFormatError(f"this task is wrong format; skip the task in {path}," " index: {i}")
                except Exception:
                    logger.exception("error while loading the task at {} (index={})".format(path, i))
        elif k == "tasks":
            if not isinstance(v, list):
                continue
            task_blocks, _ = get_task_blocks(task_dict_list=v)
            if task_blocks is None:
                continue
            for i, task_dict in enumerate(task_blocks):
                _i = i + pre_task_num
                try:
                    t = load_task(
                        path=path,
                        index=_i,
                        task_block_dict=task_dict,
                        role_name=role_name,
                        collection_name=collection_name,
                        collections_in_play=collections_in_play,
                        play_index=index,
                        parent_key=pbObj.key,
                        parent_local_key=pbObj.local_key,
                        yaml_lines=yaml_lines,
                        basedir=basedir,
                    )
                    tasks.append(t)
                except TaskFormatError:
                    if skip_task_format_error:
                        logger.debug("this task is wrong format; skip the task in {}," " index: {}; skip this".format(path, i))
                        continue
                    else:
                        raise TaskFormatError(f"this task is wrong format; skip the task in {path}," " index: {i}")
                except Exception:
                    logger.exception("error while loading the task at {} (index={})".format(path, i))
        elif k == "post_tasks":
            if not isinstance(v, list):
                continue
            task_blocks, _ = get_task_blocks(task_dict_list=v)
            if task_blocks is None:
                continue
            for i, task_dict in enumerate(task_blocks):
                _i = i + pre_task_num + task_num
                try:
                    t = load_task(
                        path=path,
                        index=_i,
                        task_block_dict=task_dict,
                        role_name=role_name,
                        collection_name=collection_name,
                        collections_in_play=collections_in_play,
                        play_index=index,
                        parent_key=pbObj.key,
                        parent_local_key=pbObj.local_key,
                        yaml_lines=yaml_lines,
                        basedir=basedir,
                    )
                    post_tasks.append(t)
                except TaskFormatError:
                    if skip_task_format_error:
                        logger.debug("this task is wrong format; skip the task in {}," " index: {}; skip this".format(path, i))
                        continue
                    else:
                        raise TaskFormatError(f"this task is wrong format; skip the task in {path}," " index: {i}")
                except Exception:
                    logger.exception("error while loading the task at {} (index={})".format(path, i))
        elif k == "roles":
            if not isinstance(v, list):
                continue
            for i, r_block in enumerate(v):
                r_name = ""
                role_options = {}
                if isinstance(r_block, dict):
                    r_name = r_block.get("role", "")
                    role_options = {}
                    for k, v in r_block.items():
                        role_options[k] = v
                elif isinstance(r_block, str):
                    r_name = r_block
                try:
                    rip = load_roleinplay(
                        name=r_name,
                        options=role_options,
                        defined_in=path,
                        role_index=i,
                        play_index=index,
                        role_name=role_name,
                        collection_name=collection_name,
                        collections_in_play=collections_in_play,
                        basedir=basedir,
                    )
                    roles.append(rip)
                except Exception:
                    logger.exception("error while loading the role in playbook at {}" " (play_index={}, role_index={})".format(path, pbObj.index, i))
        elif k == "import_playbook":
            if not isinstance(v, str):
                continue
            import_module = k
            import_playbook = v
        elif k == "include":
            if not isinstance(v, str):
                continue
            import_module = k
            import_playbook = v
        else:
            play_options.update({k: v})

    pbObj.name = play_name
    pbObj.defined_in = path
    pbObj.import_module = import_module
    pbObj.import_playbook = import_playbook
    pbObj.pre_tasks = pre_tasks
    pbObj.tasks = tasks
    pbObj.post_tasks = post_tasks
    pbObj.roles = roles
    pbObj.options = play_options
    pbObj.become = BecomeInfo.from_options(play_options)
    pbObj.collections_in_play = collections_in_play

    return pbObj


def load_roleinplay(
    name,
    options,
    defined_in,
    role_index,
    play_index,
    role_name="",
    collection_name="",
    collections_in_play=[],
    playbook_yaml="",
    basedir="",
):
    ripObj = RoleInPlay()
    if name == "":
        if "name" in options:
            name = options["name"]
            options.pop("name", None)
    ripObj.name = name
    ripObj.options = options
    if basedir != "":
        if defined_in.startswith(basedir):
            defined_in = defined_in[len(basedir) :]
            if defined_in.startswith("/"):
                defined_in = defined_in[1:]
    ripObj.defined_in = defined_in
    ripObj.role = role_name
    ripObj.collection = collection_name
    ripObj.role_index = role_index
    ripObj.play_index = play_index
    ripObj.collections_in_play = collections_in_play

    return ripObj


def load_playbook(path="", yaml_str="", role_name="", collection_name="", basedir="", skip_playbook_format_error=True, skip_task_format_error=True):
    pbObj = Playbook()
    fullpath = ""
    if yaml_str:
        fullpath = path
    else:
        if os.path.exists(path) and path != "" and path != ".":
            fullpath = path
        if os.path.exists(os.path.join(basedir, path)):
            fullpath = os.path.normpath(os.path.join(basedir, path))
        if fullpath == "":
            raise ValueError("file not found")
    defined_in = fullpath
    if basedir != "":
        if defined_in.startswith(basedir):
            defined_in = defined_in[len(basedir) :]
            if defined_in.startswith("/"):
                defined_in = defined_in[1:]
    pbObj.defined_in = defined_in
    pbObj.name = os.path.basename(fullpath)
    pbObj.role = role_name
    pbObj.collection = collection_name
    pbObj.set_key()
    yaml_lines = ""
    data = None
    if yaml_str:
        try:
            yaml_lines = yaml_str
            data = yaml.safe_load(yaml_lines)
        except Exception as e:
            if skip_playbook_format_error:
                logger.debug(f"failed to load this yaml string to load playbook, skip this yaml; {e}")
            else:
                raise PlaybookFormatError(f"failed to load this yaml string to load playbook; {e}")
    elif fullpath != "":
        with open(fullpath, "r") as file:
            try:
                yaml_lines = file.read()
                data = yaml.safe_load(yaml_lines)
            except Exception as e:
                if skip_playbook_format_error:
                    logger.debug(f"failed to load this yaml file to load playbook, skip this yaml; {e}")
                else:
                    raise PlaybookFormatError(f"failed to load this yaml file to load playbook; {e}")
    if data is None:
        return pbObj
    if not isinstance(data, list):
        raise PlaybookFormatError("playbook must be loaded as a list, but got {}".format(type(data).__name__))

    if yaml_lines:
        pbObj.yaml_lines = yaml_lines

    plays = []
    for i, play_dict in enumerate(data):
        try:
            play = load_play(
                path=defined_in,
                index=i,
                play_block_dict=play_dict,
                role_name=role_name,
                collection_name=collection_name,
                parent_key=pbObj.key,
                parent_local_key=pbObj.local_key,
                yaml_lines=yaml_str,
                basedir=basedir,
                skip_task_format_error=skip_task_format_error,
            )
            plays.append(play)
        except PlaybookFormatError:
            if skip_playbook_format_error:
                logger.debug("this play is wrong format; skip the play in {}, index: {}, skip this play".format(fullpath, i))
            else:
                raise PlaybookFormatError(f"this play is wrong format; skip the play in {fullpath}, index: {i}")
        except Exception:
            logger.exception("error while loading the play at {} (index={})".format(fullpath, i))
    pbObj.plays = plays

    return pbObj


def load_playbooks(path, basedir="", skip_playbook_format_error=True, skip_task_format_error=True, load_children=True):
    if path == "":
        return []
    patterns = [
        path + "/*.yml",
        path + "/*.yaml",
        path + "/playbooks/**/*.yml",
        path + "/playbooks/**/*.yaml",
    ]
    candidates = safe_glob(patterns, recursive=True)
    playbooks = []
    playbook_names = []
    for fpath in candidates:
        if could_be_playbook(fpath):
            if "/roles/" in fpath:
                continue
            p = None
            try:
                p = load_playbook(
                    path=fpath,
                    basedir=basedir,
                    skip_playbook_format_error=skip_playbook_format_error,
                    skip_task_format_error=skip_task_format_error,
                )
            except PlaybookFormatError as e:
                if skip_playbook_format_error:
                    logger.debug("this file is not in a playbook format, maybe not a playbook file, skip this: {}".format(e.args[0]))
                    continue
                else:
                    raise PlaybookFormatError(f"this file is not in a playbook format, maybe not a playbook file: {e.args[0]}")
            except Exception:
                logger.exception("error while loading the playbook at {}".format(fpath))
            if load_children:
                playbooks.append(p)
                playbook_names.append(p.defined_in)
            else:
                playbooks.append(p.defined_in)
                playbook_names.append(p.defined_in)
    if not load_children:
        playbooks = sorted(playbooks)
    return playbooks


def load_role(
    path,
    name="",
    collection_name="",
    module_dir_paths=[],
    basedir="",
    use_ansible_doc=True,
    skip_playbook_format_error=True,
    skip_task_format_error=True,
    load_children=True,
):
    roleObj = Role()
    fullpath = ""
    if os.path.exists(path) and path != "" and path != ".":
        fullpath = path
    if os.path.exists(os.path.join(basedir, path)):
        fullpath = os.path.normpath(os.path.join(basedir, path))
    if fullpath == "":
        raise ValueError(f"directory not found: {path}, {basedir}")
    meta_file_path = ""
    defaults_dir_path = ""
    vars_dir_path = ""
    tasks_dir_path = ""
    handlers_dir_path = ""
    includes_dir_path = ""
    if fullpath != "":
        meta_file_path = os.path.join(fullpath, "meta/main.yml")
        defaults_dir_path = os.path.join(fullpath, "defaults")
        vars_dir_path = os.path.join(fullpath, "vars")
        tasks_dir_path = os.path.join(fullpath, "tasks")
        handlers_dir_path = os.path.join(fullpath, "handlers")
        includes_dir_path = os.path.join(fullpath, "includes")
    if os.path.exists(meta_file_path):
        with open(meta_file_path, "r") as file:
            try:
                roleObj.metadata = yaml.safe_load(file)
            except Exception as e:
                logger.debug("failed to load this yaml file to raed metadata; {}".format(e.args[0]))

            if roleObj.metadata is not None and isinstance(roleObj.metadata, dict):
                roleObj.dependency["roles"] = roleObj.metadata.get("dependencies", [])
                roleObj.dependency["collections"] = roleObj.metadata.get("collections", [])

    requirements_yml_path = os.path.join(fullpath, "requirements.yml")
    if os.path.exists(requirements_yml_path):
        with open(requirements_yml_path, "r") as file:
            try:
                roleObj.requirements = yaml.safe_load(file)
            except Exception as e:
                logger.debug("failed to load requirements.yml; {}".format(e.args[0]))

    parts = tasks_dir_path.split("/")
    if len(parts) < 2:
        raise ValueError("role path is wrong")
    role_name = parts[-2] if name == "" else name
    roleObj.name = role_name
    defined_in = fullpath
    if basedir != "":
        if defined_in.startswith(basedir):
            defined_in = defined_in[len(basedir) :]
            if defined_in.startswith("/"):
                defined_in = defined_in[1:]
    roleObj.defined_in = defined_in
    collection = ""
    fqcn = role_name
    if collection_name != "":
        collection = collection_name
        fqcn = "{}.{}".format(collection_name, role_name)
    roleObj.collection = collection
    roleObj.fqcn = fqcn
    roleObj.set_key()

    if os.path.exists(os.path.join(fullpath, "playbooks")):
        playbook_files = safe_glob(fullpath + "/playbooks/**/*.yml", recursive=True)
        playbooks = []
        for f in playbook_files:
            p = None
            try:
                p = load_playbook(
                    f,
                    role_name=role_name,
                    collection_name=collection_name,
                    basedir=basedir,
                    skip_playbook_format_error=skip_playbook_format_error,
                    skip_task_format_error=skip_task_format_error,
                )
            except PlaybookFormatError as e:
                if skip_playbook_format_error:
                    logger.debug("this file is not in a playbook format, maybe not a playbook file, skip this: {}".format(e.args[0]))
                    continue
                else:
                    raise PlaybookFormatError(f"this file is not in a playbook format, maybe not a playbook file: {e.args[0]}")
            except Exception:
                logger.exception("error while loading the playbook at {}".format(f))
            if load_children:
                playbooks.append(p)
            else:
                playbooks.append(p.defined_in)
        if not load_children:
            playbooks = sorted(playbooks)
        roleObj.playbooks = playbooks

    if os.path.exists(defaults_dir_path):
        patterns = [
            defaults_dir_path + "/**/*.yml",
            defaults_dir_path + "/**/*.yaml",
        ]
        defaults_yaml_files = safe_glob(patterns, recursive=True)
        default_variables = {}
        for fpath in defaults_yaml_files:
            with open(fpath, "r") as file:
                try:
                    vars_in_yaml = yaml.safe_load(file)
                    if vars_in_yaml is None:
                        continue
                    if not isinstance(vars_in_yaml, dict):
                        continue
                    default_variables.update(vars_in_yaml)
                except Exception as e:
                    logger.debug("failed to load this yaml file to read default" " variables; {}".format(e.args[0]))
        roleObj.default_variables = default_variables

    if os.path.exists(vars_dir_path):
        patterns = [vars_dir_path + "/**/*.yml", vars_dir_path + "/**/*.yaml"]
        vars_yaml_files = safe_glob(patterns, recursive=True)
        variables = {}
        for fpath in vars_yaml_files:
            with open(fpath, "r") as file:
                try:
                    vars_in_yaml = yaml.safe_load(file)
                    if vars_in_yaml is None:
                        continue
                    if not isinstance(vars_in_yaml, dict):
                        continue
                    variables.update(vars_in_yaml)
                except Exception as e:
                    logger.debug("failed to load this yaml file to read variables; {}".format(e.args[0]))
        roleObj.variables = variables

    modules = []
    module_files = search_module_files(fullpath, module_dir_paths)

    if not load_children:
        use_ansible_doc = False

    module_specs = {}
    if use_ansible_doc:
        module_specs = get_module_specs_by_ansible_doc(
            module_files=module_files,
            fqcn_prefix=collection_name,
            search_path=fullpath,
        )

    for module_file_path in module_files:
        m = None
        try:
            m = load_module(
                module_file_path,
                collection_name=collection_name,
                role_name=fqcn,
                basedir=basedir,
                use_ansible_doc=use_ansible_doc,
                module_specs=module_specs,
            )
        except Exception:
            logger.exception("error while loading the module at {}".format(module_file_path))
        if load_children:
            modules.append(m)
        else:
            modules.append(m.defined_in)
    if not load_children:
        modules = sorted(modules)
    roleObj.modules = modules

    patterns = [tasks_dir_path + "/**/*.yml", tasks_dir_path + "/**/*.yaml"]
    # ansible.network collection has this type of another taskfile directory
    if os.path.exists(includes_dir_path):
        patterns.extend(
            [
                includes_dir_path + "/**/*.yml",
                includes_dir_path + "/**/*.yaml",
            ]
        )
    if os.path.exists(handlers_dir_path):
        patterns.extend(
            [
                handlers_dir_path + "/**/*.yml",
                handlers_dir_path + "/**/*.yaml",
            ]
        )
    task_yaml_files = safe_glob(patterns, recursive=True)

    taskfiles = []
    for task_yaml_path in task_yaml_files:
        try:
            tf = load_taskfile(
                task_yaml_path,
                role_name=fqcn,
                collection_name=collection_name,
                basedir=basedir,
                skip_task_format_error=skip_task_format_error,
            )
        except TaskFormatError as e:
            if skip_task_format_error:
                logger.debug(f"Task format error found; skip this taskfile {task_yaml_path}")
            else:
                raise TaskFormatError(f"Task format error found: {e.args[0]}")
        except Exception:
            logger.exception("error while loading the task file at {}".format(task_yaml_path))
        if load_children:
            taskfiles.append(tf)
        else:
            taskfiles.append(tf.defined_in)
    if not load_children:
        taskfiles = sorted(taskfiles)
    roleObj.taskfiles = taskfiles

    return roleObj


def load_roles(path, basedir="", use_ansible_doc=True, skip_playbook_format_error=True, skip_task_format_error=True, load_children=True):

    if path == "":
        return []
    roles_patterns = ["roles", "playbooks/roles", "playbook/roles"]
    roles_dir_path = ""
    for r_p in roles_patterns:
        candidate = os.path.join(path, r_p)
        if os.path.exists(candidate):
            roles_dir_path = candidate
            break
    if roles_dir_path == "":
        return []
    dirs = os.listdir(roles_dir_path)
    roles = []
    for dir_name in dirs:
        role_dir = os.path.join(roles_dir_path, dir_name)
        try:
            r = load_role(
                role_dir,
                basedir=basedir,
                use_ansible_doc=use_ansible_doc,
                skip_playbook_format_error=skip_playbook_format_error,
                skip_task_format_error=skip_task_format_error,
            )
        except Exception:
            logger.exception("error while loading the role at {}".format(role_dir))
        if load_children:
            roles.append(r)
        else:
            roles.append(r.defined_in)
    if not load_children:
        roles = sorted(roles)
    return roles


def load_requirements(path):
    requirements = {}
    requirements_yml_path = os.path.join(path, "requirements.yml")
    if os.path.exists(requirements_yml_path):
        with open(requirements_yml_path, "r") as file:
            try:
                requirements = yaml.safe_load(file)
            except Exception as e:
                logger.debug("failed to load requirements.yml; {}".format(e.args[0]))
    return requirements


def load_installed_roles(installed_roles_path):
    search_path = installed_roles_path
    if installed_roles_path == "" or not os.path.exists(search_path):
        return []
    dirs = os.listdir(search_path)
    roles = []
    basedir = os.path.dirname(os.path.normpath(installed_roles_path))
    for d in dirs:
        role_path = os.path.join(installed_roles_path, d)
        role_meta_files = safe_glob(role_path + "/**/meta/main.yml", recursive=True)

        roles_root_dirs = set([f.split("/roles/")[-2] for f in role_meta_files if "/roles/" in f])
        module_dirs = []
        for role_root_dir in roles_root_dirs:
            moddirs = find_module_dirs(role_root_dir)
            module_dirs.extend(moddirs)

        for i, role_meta_file in enumerate(role_meta_files):
            role_dir_path = role_meta_file.replace("/meta/main.yml", "")
            module_dir_paths = []
            if i == 0:
                module_dir_paths = module_dirs
            try:
                r = load_role(
                    role_dir_path,
                    module_dir_paths=module_dir_paths,
                    basedir=basedir,
                )
                roles.append(r)
            except Exception:
                logger.exception("error while loading the role at {}".format(role_dir_path))
    return roles


def load_module(module_file_path, collection_name="", role_name="", basedir="", use_ansible_doc=True, module_specs={}):
    moduleObj = Module()
    if module_file_path == "":
        raise ValueError("require module file path to load a Module")
    fullpath = ""
    if os.path.exists(module_file_path) and module_file_path != "" and module_file_path != ".":
        fullpath = module_file_path
    if os.path.exists(os.path.join(basedir, module_file_path)):
        fullpath = os.path.normpath(os.path.join(basedir, module_file_path))
    if fullpath == "":
        raise ValueError(f"module file not found: {module_file_path}, {basedir}")

    file_name = os.path.basename(module_file_path)
    module_name = file_name.replace(".py", "")

    # some collections have modules like `plugins/modules/xxxx/yyyy.py`
    # so try finding `xxxx` part by checking module file path
    for dir_pattern in module_dir_patterns:
        separator = dir_pattern + "/"
        if separator in module_file_path:
            module_name = module_file_path.split(separator)[-1].replace(".py", "").replace("/", ".")
            break

    moduleObj.name = module_name
    if collection_name != "":
        moduleObj.collection = collection_name
        moduleObj.fqcn = "{}.{}".format(collection_name, module_name)
    elif role_name != "":
        # if module is defined in a role, it does not have real fqcn
        moduleObj.role = role_name
        moduleObj.fqcn = module_name
    defined_in = module_file_path
    if basedir != "":
        if defined_in.startswith(basedir):
            defined_in = defined_in[len(basedir) :]
            if defined_in.startswith("/"):
                defined_in = defined_in[1:]
    moduleObj.defined_in = defined_in

    arguments = []
    doc_yaml = ""
    examples = ""
    if use_ansible_doc:
        # running `ansible-doc` for each module causes speed problem due to overhead,
        # so use it for all modules and pick up the doc for the module here
        if module_specs:
            doc_yaml = module_specs.get(moduleObj.fqcn, {}).get("doc", "")
            examples = module_specs.get(moduleObj.fqcn, {}).get("examples", "")
    else:
        # parse the script file for a quick scan (this does not contain doc from `doc_fragments`)
        doc_yaml = get_documentation_in_module_file(fullpath)
    if doc_yaml:
        doc_dict = {}
        try:
            doc_dict = yaml.safe_load(doc_yaml)
        except Exception:
            logger.debug(f"failed to load the arguments documentation of the module: {module_name}")
        arg_specs = doc_dict.get("options", {})
        for arg_name in arg_specs:
            arg_spec = arg_specs[arg_name]
            arg_value_type = get_class_by_arg_type(arg_spec.get("type", None))
            arg_value_type_str = ""
            if arg_value_type:
                arg_value_type_str = arg_value_type.__name__
            arg = ModuleArgument(
                name=arg_name,
                type=arg_value_type_str,
                required=boolean(arg_spec.get("required", "false")),
                description=arg_spec.get("description", ""),
                default=arg_spec.get("default", None),
                choices=arg_spec.get("choices", None),
                aliases=arg_spec.get("aliases", None),
            )
            arguments.append(arg)
    moduleObj.documentation = doc_yaml
    moduleObj.examples = examples
    moduleObj.arguments = arguments

    moduleObj.set_key()

    return moduleObj


builtin_modules_file_name = "ansible_builtin_modules.json"


def load_builtin_modules():
    base_path = os.path.dirname(__file__)
    data_path = os.path.join(base_path, builtin_modules_file_name)
    module_list = ObjectList.from_json(fpath=data_path)
    module_dict = {m.name: m for m in module_list.items}
    return module_dict


# modules in a SCM repo should be in `library` dir in the best practice case
# https://docs.ansible.com/ansible/2.8/user_guide/playbooks_best_practices.html
# however, it is often defined in `plugins/modules` directory,
# so we search both the directories
def load_modules(path, basedir="", collection_name="", module_dir_paths=[], use_ansible_doc=True, load_children=True):
    if path == "":
        return []
    if not os.path.exists(path):
        return []
    module_files = search_module_files(path, module_dir_paths)

    if len(module_files) == 0:
        return []

    if not load_children:
        use_ansible_doc = False

    module_specs = {}
    if use_ansible_doc:
        module_specs = get_module_specs_by_ansible_doc(
            module_files=module_files,
            fqcn_prefix=collection_name,
            search_path=path,
        )

    modules = []
    for module_file_path in module_files:
        m = None
        try:
            m = load_module(
                module_file_path,
                collection_name=collection_name,
                basedir=basedir,
                use_ansible_doc=use_ansible_doc,
                module_specs=module_specs,
            )
        except Exception:
            logger.exception("error while loading the module at {}".format(module_file_path))
        if load_children:
            modules.append(m)
        else:
            modules.append(m.defined_in)
    if not load_children:
        modules = sorted(modules)
    return modules


def load_task(
    path,
    index,
    task_block_dict,
    role_name="",
    collection_name="",
    collections_in_play=[],
    play_index=-1,
    parent_key="",
    parent_local_key="",
    yaml_lines="",
    basedir="",
):

    taskObj = Task()
    fullpath = ""
    if yaml_lines:
        fullpath = path
    else:
        if os.path.exists(path) and path != "" and path != ".":
            fullpath = path
        if os.path.exists(os.path.join(basedir, path)):
            fullpath = os.path.normpath(os.path.join(basedir, path))
        if fullpath == "":
            raise ValueError("file not found")
        if not fullpath.endswith(".yml") and not fullpath.endswith(".yaml"):
            raise ValueError('task yaml file must be ".yml" or ".yaml"')
    if task_block_dict is None:
        raise ValueError("task block dict is required to load Task")
    if not isinstance(task_block_dict, dict):
        raise TaskFormatError(f"this task block is not a dict, but {type(task_block_dict).__name__}; maybe this is not a task")
    data_block = task_block_dict
    task_name = ""
    module_name = find_module_name(task_block_dict)
    module_short_name = module_name.split(".")[-1]
    task_options = {}
    module_options = {}
    for k, v in data_block.items():
        if k == "name":
            task_name = v
        if k == module_name:
            module_options = v
        else:
            task_options.update({k: v})

    taskObj.set_yaml_lines(fullpath=fullpath, yaml_lines=yaml_lines, task_name=task_name, module_name=module_name, module_options=module_options)

    # module_options can be passed as a string like below
    #
    # - name: sample of string module options
    #   ufw: port={{ item }} proto=tcp rule=allow
    #   with_items:
    #   - 5222
    #   - 5269
    if isinstance(module_options, str):
        if string_module_options_re.match(module_options):
            new_module_options = {}
            unknown_key = "__unknown_option_name__"
            if module_short_name in ["import_role", "include_role"]:
                unknown_key = "name"
            elif module_short_name in [
                "import_tasks",
                "include_tasks",
                "include",
            ]:
                unknown_key = "file"
            matched_options = string_module_options_re.findall(module_options)
            if len(matched_options) == 0:
                new_module_options[unknown_key] = module_options
            else:
                unknown_key_val = module_options.split(matched_options[0])[0]
                if unknown_key_val != "":
                    new_module_options[unknown_key] = unknown_key_val
                for p in matched_options:
                    key = p.split("=")[0]
                    val = "=".join(p.split("=")[1:])
                    new_module_options[key] = val
            module_options = new_module_options
    executable = module_name
    executable_type = ExecutableType.MODULE_TYPE
    if module_short_name in ["import_role", "include_role"]:
        role_ref = ""
        if isinstance(module_options, str):
            role_ref = module_options
        elif isinstance(module_options, dict):
            role_ref = module_options.get("name", "")
        executable = role_ref
        executable_type = ExecutableType.ROLE_TYPE
    if module_short_name in ["import_tasks", "include_tasks", "include"]:
        taskfile_ref = ""
        if isinstance(module_options, str):
            taskfile_ref = module_options
        elif isinstance(module_options, dict):
            taskfile_ref = module_options.get("file", "")
        executable = taskfile_ref
        executable_type = ExecutableType.TASKFILE_TYPE

    taskObj.name = task_name
    taskObj.role = role_name
    taskObj.collection = collection_name
    defined_in = fullpath
    if basedir != "":
        if defined_in.startswith(basedir):
            defined_in = defined_in[len(basedir) :]
            if defined_in.startswith("/"):
                defined_in = defined_in[1:]
    taskObj.defined_in = defined_in
    taskObj.index = index
    taskObj.play_index = play_index
    taskObj.executable = executable
    taskObj.executable_type = executable_type
    taskObj.collections_in_play = collections_in_play
    taskObj.set_key(parent_key, parent_local_key)

    variables = {}
    # get variables for this task
    if "vars" in task_options:
        vars_in_task = task_options.get("vars", {})
        if vars_in_task is not None and isinstance(vars_in_task, dict):
            variables.update(vars_in_task)

    set_facts = {}
    # if the Task is set_fact, set variables too
    if module_short_name == "set_fact":
        if isinstance(module_options, dict):
            set_facts.update(module_options)

    registered_variables = {}
    # set variables if this task register a new var
    if "register" in task_options:
        register_var_name = task_options.get("register", "")
        if register_var_name is not None and isinstance(register_var_name, str) and register_var_name != "":
            registered_variables.update({register_var_name: taskObj.key})

    # set loop variables when loop / with_xxxx are there
    loop_info = {}
    for k in task_options:
        if k in loop_task_option_names:
            loop_var = task_options.get("loop_control", {}).get("loop_var", "item")
            loop_info[loop_var] = task_options.get(k, [])

    taskObj.options = task_options
    taskObj.become = BecomeInfo.from_options(task_options)
    taskObj.variables = variables
    taskObj.registered_variables = registered_variables
    taskObj.set_facts = set_facts
    taskObj.loop = loop_info
    taskObj.module = module_name
    taskObj.module_options = module_options

    return taskObj


def load_taskfile(path, yaml_str="", role_name="", collection_name="", basedir="", skip_task_format_error=True):
    tfObj = TaskFile()
    fullpath = ""
    if yaml_str:
        fullpath = path
    else:
        if os.path.exists(path) and path != "" and path != ".":
            fullpath = path
        if os.path.exists(os.path.join(basedir, path)):
            fullpath = os.path.normpath(os.path.join(basedir, path))
        if fullpath == "":
            raise ValueError("file not found")
        if not fullpath.endswith(".yml") and not fullpath.endswith(".yaml"):
            raise ValueError('task yaml file must be ".yml" or ".yaml"')
    tfObj.name = os.path.basename(fullpath)
    defined_in = fullpath
    if basedir != "":
        if defined_in.startswith(basedir):
            defined_in = defined_in[len(basedir) :]
            if defined_in.startswith("/"):
                defined_in = defined_in[1:]
    tfObj.defined_in = defined_in
    if role_name != "":
        tfObj.role = role_name
    if collection_name != "":
        tfObj.collection = collection_name
    tfObj.set_key()

    task_dicts, yaml_lines = get_task_blocks(fpath=fullpath, yaml_str=yaml_str)

    if yaml_str and not yaml_lines:
        yaml_lines = yaml_str
    tfObj.yaml_lines = yaml_lines

    if task_dicts is None:
        return tfObj
    tasks = []
    for i, t_dict in enumerate(task_dicts):
        try:
            t = load_task(
                fullpath,
                i,
                t_dict,
                role_name,
                collection_name,
                yaml_lines=yaml_lines,
                parent_key=tfObj.key,
                parent_local_key=tfObj.local_key,
                basedir=basedir,
            )
            tasks.append(t)
        except TaskFormatError:
            if skip_task_format_error:
                logger.debug("this task is wrong format; skip the task in {}, index: {}; skip this".format(fullpath, i))
                continue
            else:
                raise TaskFormatError(f"Task format error found; {fullpath}, index: {i}")
        except Exception:
            logger.exception("error while loading the task at {}, index: {}".format(fullpath, i))
    tfObj.tasks = tasks

    return tfObj


# playbooks possibly include/import task files around the playbook file
# we search this type of isolated taskfile in `playbooks` and `tasks` dir
def load_taskfiles(path, basedir="", load_children=True):
    if not os.path.exists(path):
        return []

    taskfile_paths = search_taskfiles_for_playbooks(path)
    if len(taskfile_paths) == 0:
        return []

    taskfiles = []
    for taskfile_path in taskfile_paths:
        try:
            tf = load_taskfile(taskfile_path, basedir=basedir)
        except Exception:
            logger.exception("error while loading the task file at {}".format(taskfile_path))
        if load_children:
            taskfiles.append(tf)
        else:
            taskfiles.append(tf.defined_in)
    if not load_children:
        taskfiles = sorted(taskfiles)
    return taskfiles


def load_collection(
    collection_dir, basedir="", use_ansible_doc=True, skip_playbook_format_error=True, skip_task_format_error=True, load_children=True
):

    colObj = Collection()
    fullpath = ""
    if os.path.exists(collection_dir):
        fullpath = collection_dir
    if os.path.exists(os.path.join(basedir, collection_dir)):
        fullpath = os.path.join(basedir, collection_dir)
    if fullpath == "":
        raise ValueError("directory not found")
    parts = fullpath.split("/")
    if len(parts) < 2:
        raise ValueError("collection directory path is wrong")
    collection_name = "{}.{}".format(parts[-2], parts[-1])

    manifest_file_path = os.path.join(fullpath, "MANIFEST.json")
    if os.path.exists(manifest_file_path):
        with open(manifest_file_path, "r") as file:
            colObj.metadata = json.load(file)

        if colObj.metadata is not None and isinstance(colObj.metadata, dict):
            colObj.dependency["collections"] = colObj.metadata.get("dependencies", {})

    files_file_path = os.path.join(fullpath, "FILES.json")
    if os.path.exists(files_file_path):
        with open(files_file_path, "r") as file:
            colObj.files = json.load(file)

    requirements_yml_path = os.path.join(fullpath, "requirements.yml")
    if os.path.exists(requirements_yml_path):
        with open(requirements_yml_path, "r") as file:
            try:
                colObj.requirements = yaml.safe_load(file)
            except Exception as e:
                logger.debug("failed to load requirements.yml; {}".format(e.args[0]))

    playbook_path_patterns = [
        fullpath + "/playbooks/**/*.yml",
        fullpath + "/playbooks/**/*.yaml",
    ]
    playbook_files = safe_glob(playbook_path_patterns, recursive=True)
    playbooks = []
    for f in playbook_files:
        p = None
        try:
            p = load_playbook(
                path=f,
                collection_name=collection_name,
                basedir=basedir,
                skip_playbook_format_error=skip_playbook_format_error,
                skip_task_format_error=skip_task_format_error,
            )
        except PlaybookFormatError as e:
            if skip_playbook_format_error:
                logger.debug("this file is not in a playbook format, maybe not a playbook file, skip this: {}".format(e.args[0]))
                continue
            else:
                raise PlaybookFormatError(f"this file is not in a playbook format, maybe not a playbook file: {e.args[0]}")
        except Exception:
            logger.exception("error while loading the playbook at {}".format(f))
            continue
        if load_children:
            playbooks.append(p)
        else:
            playbooks.append(p.defined_in)
    if not load_children:
        playbooks = sorted(playbooks)

    taskfile_paths = search_taskfiles_for_playbooks(fullpath)
    if len(taskfile_paths) > 0:
        taskfiles = []
        for taskfile_path in taskfile_paths:
            try:
                tf = load_taskfile(taskfile_path, basedir=basedir)
            except Exception:
                logger.exception("error while loading the task file at {}".format(taskfile_path))
                continue
            if load_children:
                taskfiles.append(tf)
            else:
                taskfiles.append(tf.defined_in)
        if not load_children:
            taskfiles = sorted(taskfiles)
        colObj.taskfiles = taskfiles

    roles = load_roles(fullpath, basedir=basedir, use_ansible_doc=use_ansible_doc, load_children=load_children)

    module_files = search_module_files(fullpath)

    modules = []
    if not load_children:
        use_ansible_doc = False

    module_specs = {}
    if use_ansible_doc:
        module_specs = get_module_specs_by_ansible_doc(
            module_files=module_files,
            fqcn_prefix=collection_name,
            search_path=fullpath,
        )

    for f in module_files:
        m = None
        try:
            m = load_module(f, collection_name=collection_name, basedir=basedir, use_ansible_doc=use_ansible_doc, module_specs=module_specs)
        except Exception:
            logger.exception("error while loading the module at {}".format(f))
            continue
        if load_children:
            modules.append(m)
        else:
            modules.append(m.defined_in)
    if not load_children:
        modules = sorted(modules)
    colObj.name = collection_name
    path = collection_dir
    if basedir != "":
        if path.startswith(basedir):
            path = path[len(basedir) :]
    colObj.path = path
    colObj.playbooks = playbooks
    colObj.roles = roles
    colObj.modules = modules

    return colObj


def load_object(loadObj):
    target_type = loadObj.target_type
    path = loadObj.path
    obj = None
    if target_type == LoadType.COLLECTION:
        obj = load_collection(collection_dir=path, basedir=path, load_children=False)
    elif target_type == LoadType.ROLE:
        obj = load_role(path=path, basedir=path, load_children=False)
    elif target_type == LoadType.PLAYBOOK:
        basedir = ""
        target_playbook_path = ""
        if loadObj.playbook_yaml:
            target_playbook_path = path
        else:
            basedir, target_playbook_path = split_target_playbook_fullpath(path)
        if loadObj.playbook_only:
            obj = load_playbook(path=target_playbook_path, yaml_str=loadObj.playbook_yaml, basedir=basedir)
        else:
            obj = load_repository(path=basedir, basedir=basedir, target_playbook_path=target_playbook_path, load_children=False)
    elif target_type == LoadType.TASKFILE:
        basedir = ""
        target_taskfile_path = ""
        if loadObj.taskfile_yaml:
            target_taskfile_path = path
        else:
            basedir, target_taskfile_path = split_target_taskfile_fullpath(path)
        if loadObj.taskfile_only:
            obj = load_taskfile(path=target_taskfile_path, yaml_str=loadObj.taskfile_yaml, basedir=basedir)
        else:
            obj = load_repository(path=basedir, basedir=basedir, target_taskfile_path=target_taskfile_path, load_children=False)
    elif target_type == LoadType.PROJECT:
        obj = load_repository(path=path, basedir=path, load_children=False)

    if hasattr(obj, "roles"):
        loadObj.roles = obj.roles
    if hasattr(obj, "playbooks"):
        loadObj.playbooks = obj.playbooks
    if hasattr(obj, "taskfiles"):
        loadObj.taskfiles = obj.taskfiles
    if hasattr(obj, "modules"):
        loadObj.modules = obj.modules

    if target_type == LoadType.ROLE:
        loadObj.roles = [obj.defined_in]
    elif target_type == LoadType.PLAYBOOK and loadObj.playbook_only:
        loadObj.playbooks = [obj.defined_in]
    elif target_type == LoadType.TASKFILE and loadObj.taskfile_only:
        loadObj.taskfiles = [obj.defined_in]

    loadObj.timestamp = datetime.datetime.utcnow().isoformat()


def find_playbook_role_module(path, use_ansible_doc=True):
    playbooks = load_playbooks(path, basedir=path, load_children=False)
    root_role = None
    try:
        root_role = load_role(path, basedir=path, use_ansible_doc=use_ansible_doc, load_children=False)
    except Exception:
        pass
    sub_roles = load_roles(path, basedir=path, use_ansible_doc=use_ansible_doc, load_children=False)
    roles = []
    if root_role and root_role.metadata:
        roles.append(".")
    if len(sub_roles) > 0:
        roles.extend(sub_roles)
    modules = load_modules(path, basedir=path, use_ansible_doc=use_ansible_doc, load_children=False)
    return playbooks, roles, modules



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

from ansible_risk_insight.models import (
    AnsibleRunContext,
    RunTargetType,
    Rule,
    Severity,
    RuleTag as Tag,
    ArgumentsType,
    ExecutableType,
    VariableType,
)


@dataclass
class ModuleArgumentValueValidationRule(Rule):
    rule_id: str = "P003"
    description: str = "Validate module argument values and set annotations"
    enabled: bool = True
    name: str = "ModuleArgumentValueValidation"
    version: str = "v0.0.1"
    severity: Severity = Severity.NONE
    tags: tuple = Tag.QUALITY
    precedence: int = 0

    def match(self, ctx: AnsibleRunContext) -> bool:
        return ctx.current.type == RunTargetType.Task

    def process(self, ctx: AnsibleRunContext):
        task = ctx.current

        if task.spec.executable_type == ExecutableType.MODULE_TYPE and task.module and task.module.arguments:

            wrong_values = []
            undefined_values = []
            unknown_type_values = []
            if task.args.type == ArgumentsType.DICT:
                for key in task.args.raw:
                    raw_value = task.args.raw[key]
                    resolved_value = None
                    if len(task.args.templated) == 1:
                        resolved_value = task.args.templated[0][key]
                    elif len(task.args.templated) > 1:
                        resolved_value = [t[key] for t in task.args.templated]
                    spec = None
                    for arg_spec in task.module.arguments:
                        if key == arg_spec.name or (arg_spec.aliases and key in arg_spec.aliases):
                            spec = arg_spec
                            break
                    if not spec:
                        continue

                    d = {"key": key}
                    wrong_val = False
                    unknown_type_val = False
                    if spec.type:
                        if not isinstance(raw_value, str) or "{{" not in raw_value:
                            if type(raw_value).__name__ != spec.type:
                                d["expected_type"] = spec.type
                                d["actual_type"] = type(raw_value).__name__
                                d["actual_value"] = raw_value
                                wrong_val = True
                        else:
                            if isinstance(resolved_value, str) and "{{" in resolved_value:
                                d["expected_type"] = spec.type
                                d["unknown_type_value"] = resolved_value
                                unknown_type_val = True
                            else:
                                if type(resolved_value).__name__ != spec.type:
                                    d["expected_type"] = spec.type
                                    d["actual_type"] = type(raw_value).__name__
                                    d["actual_value"] = raw_value
                                    wrong_val = True

                    if wrong_val:
                        wrong_values.append(d)

                    if unknown_type_val:
                        unknown_type_values.append(d)

                    sub_args = task.args.get(key)
                    if sub_args:
                        undefined_vars = [v.name for v in sub_args.vars if v and v.type == VariableType.Unknown]
                        if undefined_vars:
                            undefined_values.append({"key": key, "value": raw_value, "undefined_variables": undefined_vars})

            task.set_annotation("module.wrong_arg_values", wrong_values, rule_id=self.rule_id)
            task.set_annotation("module.undefined_values", undefined_values, rule_id=self.rule_id)
            task.set_annotation("module.unknown_type_values", unknown_type_values, rule_id=self.rule_id)

        # TODO: find duplicate keys

        return None
